"""apport.PackageInfo class implementation for python-apt and dpkg.

This is used on Debian and derivatives such as Ubuntu.
"""

# Copyright (C) 2007 - 2011 Canonical Ltd.
# Author: Martin Pitt <martin.pitt@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

# pylint: disable=too-many-lines
# TODO: Address following pylint complaints
# pylint: disable=invalid-name

# From Python 3.12 on, it doesn't try to evaluate type signatures at runtime
# anymore. This behaviour is helpful to us as apt_sl.Deb822SourceEntry isn't
# always a valid type, so use the __future__ mechanism to have this behaviour
# in older versions as well.
from __future__ import annotations

import contextlib
import datetime
import functools
import glob
import gzip
import hashlib
import json
import logging
import os
import pathlib
import pickle
import platform
import re
import shutil
import stat
import subprocess
import sys
import tempfile
import time
import urllib.error
import urllib.parse
import urllib.request
from collections.abc import Iterator

import apt
import apt_pkg
import aptsources.sourceslist as apt_sl
from aptsources.sourceslist import SourceEntry

import apport.logging
from apport.packaging import PackageInfo

WITH_DEB822_SUPPORT = hasattr(apt_sl, "Deb822SourceEntry")

_ARCH_MULTIARCH_TRIPLET_MAPPING = {
    "amd64": "x86_64-linux-gnu",
    "arm64": "aarch64-linux-gnu",
    "armhf": "arm-linux-gnueabihf",
    "i386": "i386-linux-gnu",
    "ppc64el": "powerpc64le-linux-gnu",
    "riscv64": "riscv64-linux-gnu",
    "s390x": "s390x-linux-gnu",
}
# The Contents-*.gz files are huge. Loading all data into memory would result in a
# dictionary with millions of entries consuming several gigabytes of memory.
# Therefore exclude unneeded paths with 100k entries or more.
# Data from 2026-04-23 on amd64:
#
# | Path                       |      jammy |      noble |   questing |   resolute |
# |----------------------------|------------|------------|------------|------------|
# | lib/modules                | 10,206,350 |  5,492,830 |    475,055 |      9,120 |
# | usr/include                |    278,938 |    271,587 |    295,470 |    298,389 |
# | usr/src                    | 39,533,381 | 21,538,896 |  2,201,315 |    420,916 |
# | usr/[^/]+/include          |    107,568 |    170,800 |    103,340 |    127,294 |
# | usr/lib/modules            |          0 |          0 |          0 |     73,589 |
# | usr/share/cargo/registry   |     28,202 |     74,721 |    102,428 |    127,537 |
# | usr/share/doc              |  3,025,577 |  2,938,577 |  2,636,718 |  2,766,784 |
# | usr/share/gimp/.../help    |     38,220 |     73,845 |     78,120 |     85,216 |
# | usr/share/gocode           |    112,802 |    168,941 |    199,405 |    227,626 |
# | usr/share/help             |    121,470 |    124,824 |    118,948 |    119,346 |
# | usr/share/icons            |    605,862 |    659,375 |    715,984 |    753,986 |
# | usr/share/libreoffice/help |     93,359 |     98,690 |    100,540 |    100,580 |
# | usr/share/locale           |     95,312 |    110,122 |    123,230 |    123,842 |
# | usr/share/man              |    149,353 |    165,310 |    175,262 |    177,041 |
# | usr/share/texlive          |    145,420 |    168,740 |    170,265 |    171,546 |
# | **remaining paths**        |  3,365,393 |  3,839,911 |  3,725,751 |  3,909,851 |
_EXCLUDED_PATHS = (
    "^(lib/modules|usr/(include|src|[^/]+/include|lib/modules|share/"
    "(cargo/registry|doc|gimp/.../help|gocode|help"
    "|icons|libreoffice/help|locale|man|texlive)))/"
)


# pylint: disable-next=no-member
def _parse_deb822_sources(source: str) -> list[apt_sl.Deb822SourceEntry]:
    if not WITH_DEB822_SUPPORT:
        return []

    sections = []
    current = []
    with open(source, encoding="utf-8") as f:
        for line in f.read().split("\n"):
            line = line.rstrip()
            if line:
                if not line.lstrip().startswith("#"):
                    current.append(line)
            else:
                if not current:
                    continue
                # pylint: disable-next=no-member
                sections.append(apt_sl.Deb822SourceEntry("\n".join(current), source))
                current = []

        if current:
            # pylint: disable-next=no-member
            sections.append(apt_sl.Deb822SourceEntry("\n".join(current), source))
        return sections


def _order_sources(apt_sources: list[str]) -> Iterator[str]:
    """Order APT sources files heuristically to get the primary sources first."""
    sources = set(apt_sources)
    info = platform.freedesktop_os_release()
    expected_primary = f"{info['ID']}.sources"
    try:
        sources.remove(expected_primary)
        yield expected_primary
    except KeyError:
        pass
    yield from sorted(sources)


# pylint: disable-next=no-member
def _load_all_sources(apt_dir: str) -> list[apt_sl.Deb822SourceEntry | SourceEntry]:
    """Given an APT configuration directory (e.g. /etc/apt/), loads up all
    the source data in one easy-to-consume list.
    """
    sources_d = os.path.join(apt_dir, "sources.list.d")
    if os.path.exists(sources_d):
        to_inspect = [
            os.path.join(sources_d, f) for f in _order_sources(os.listdir(sources_d))
        ]
    else:
        to_inspect = []

    sources_old_file = os.path.join(apt_dir, "sources.list")
    if os.path.exists(sources_old_file):
        to_inspect[0:0] = [sources_old_file]

    sources = []
    for path in to_inspect:
        if path.endswith(".list"):
            with open(path, encoding="utf-8") as f:
                sources.extend([SourceEntry(line, path) for line in f])
        elif path.endswith(".sources") and WITH_DEB822_SUPPORT:
            sources.extend(_parse_deb822_sources(path))

    return sources


def _map_mirror_to_arch(uri: str, target_arch: str) -> str:
    """Map the givin archive mirror URI to a valid one for the givin architecture.

    The Ubuntu archive is split. The primary mirrors only has packages
    for the amd64 and i386 architectures. The ports mirrors host all
    remaining architectures.
    """
    if target_arch in {"amd64", "i386"}:
        ports_match = re.match("http://([a-z.]*)ports.ubuntu.com/ubuntu-ports/?", uri)
        if ports_match:
            return f"http://{ports_match.group(1)}archive.ubuntu.com/ubuntu"
        return uri

    primary_match = re.match("http://([a-z.]*)archive.ubuntu.com/ubuntu/?$", uri)
    if primary_match:
        return f"http://{primary_match.group(1)}ports.ubuntu.com/ubuntu-ports"
    return uri


def _read_mirror_file(uri: str) -> list[str]:
    """Read an apt-transport-mirror configuration file.

    Note: The metadata will be stripped for simplicity reasons.
    """
    assert uri.startswith("mirror+file:")
    path = pathlib.Path(uri[12:])
    lines = [line.strip() for line in path.read_text("utf-8").split("\n")]
    mirrors = [
        line.split("\t", maxsplit=1)[0] for line in lines if line and line[0] != "#"
    ]
    return mirrors


def _usr_merge_alternative(path: str) -> str | None:
    """Determine /usr-merge alternative name.

    With /usr-merge some binaries appear in /usr but their .list file
    doesn't reflect that. Vice-versa is possible as well. So strip or
    add the /usr from the path. Return None in case there is no
    alternative path.
    """
    if re.match("^/usr/(bin|lib|lib32|lib64|libx32|sbin)/", path):
        return path[4:]
    if re.match("^/(bin|lib|lib32|lib64|libx32|sbin)/", path):
        return f"/usr{path}"
    return None


class __AptDpkgPackageInfo(PackageInfo):
    # pylint: disable=too-many-instance-attributes,too-many-public-methods
    """Concrete apport.packaging.PackageInfo class implementation for
    python-apt and dpkg, as found on Debian and derivatives such as Ubuntu."""

    def __init__(self):
        self._apt_cache = None
        self._current_release_codename = None
        self._sandbox_apt_cache = None
        self._sandbox_apt_cache_arch = None
        self._contents_dir = None
        self._mirror = None
        self._virtual_mapping_obj = None
        self._contents_mapping_obj = None
        self._launchpad_base = "https://api.launchpad.net/devel"
        self._contents_update = False

    def __del__(self):
        try:
            if self._contents_dir:
                shutil.rmtree(self._contents_dir)
        except AttributeError:
            pass

    def _virtual_mapping(self, configdir):
        if self._virtual_mapping_obj is not None:
            return self._virtual_mapping_obj

        mapping_file = os.path.join(configdir, "virtual_mapping.pickle")
        if os.path.exists(mapping_file):
            with open(mapping_file, "rb") as fp:
                self._virtual_mapping_obj = pickle.load(fp)
        else:
            self._virtual_mapping_obj = {}

        return self._virtual_mapping_obj

    def _save_virtual_mapping(self, configdir):
        mapping_file = os.path.join(configdir, "virtual_mapping.pickle")
        if self._virtual_mapping_obj is not None:
            with open(mapping_file, "wb") as fp:
                pickle.dump(self._virtual_mapping_obj, fp)

    def _contents_mapping(self, configdir, release, arch):
        if (
            self._contents_mapping_obj
            and self._contents_mapping_obj["release"] == release
            and self._contents_mapping_obj["arch"] == arch
        ):
            return self._contents_mapping_obj

        mapping_file = os.path.join(
            configdir, f"contents_mapping-{release}-{arch}.pickle"
        )
        if os.path.exists(mapping_file) and os.stat(mapping_file).st_size == 0:
            os.remove(mapping_file)
        if os.path.exists(mapping_file):
            with open(mapping_file, "rb") as fp:
                self._contents_mapping_obj = pickle.load(fp)
        else:
            self._contents_mapping_obj = {"release": release, "arch": arch}

        return self._contents_mapping_obj

    def _save_contents_mapping(self, configdir, release, arch):
        mapping_file = os.path.join(
            configdir, f"contents_mapping-{release}-{arch}.pickle"
        )
        if self._contents_mapping_obj is not None:
            try:
                with open(mapping_file, "wb") as fp:
                    pickle.dump(self._contents_mapping_obj, fp)
            # rather than crashing on systems with little memory just don't
            # write the crash file
            except MemoryError:
                pass

    def _clear_apt_cache(self) -> None:
        # The rootdir option to apt.Cache modifies the global state
        apt_pkg.config.clear("Dir")
        apt_pkg.init_config()
        apt_pkg.init_system()

        self._apt_cache = None
        self._sandbox_apt_cache = None

    def _cache(self):
        """Return apt.Cache() (initialized lazily)."""
        if not self._apt_cache:
            self._clear_apt_cache()
            # avoid spewage on stdout
            progress = apt.progress.base.OpProgress()
            self._apt_cache = apt.Cache(progress=progress)
        return self._apt_cache

    def _sandbox_cache(
        self,
        aptroot,
        apt_dir,
        fetchProgress,
        distro_name,
        release_codename,
        origins,
        arch,
    ):  # pylint: disable=too-many-arguments
        """Build apt sandbox and return apt.Cache(rootdir=) (initialized
        lazily).

        Clear the package selection on subsequent calls.
        """
        if not self._sandbox_apt_cache or arch != self._sandbox_apt_cache_arch:
            self._clear_apt_cache()
            self._build_apt_sandbox(
                aptroot, apt_dir, distro_name, release_codename, origins
            )
            rootdir = os.path.abspath(aptroot)
            self._sandbox_apt_cache = apt.Cache(rootdir=rootdir)
            self._sandbox_apt_cache_arch = arch
            try:
                # We don't need to update this multiple times.
                self._sandbox_apt_cache.update(fetchProgress)
            except apt.cache.FetchFailedException as error:
                raise SystemError(str(error)) from error
            self._sandbox_apt_cache.open()
        else:
            self._sandbox_apt_cache.clear()
        return self._sandbox_apt_cache

    def _apt_pkg(self, package):
        """Return apt.Cache()[package] (initialized lazily).

        Throw a ValueError if the package does not exist.
        """
        try:
            return self._cache()[package]
        except KeyError:
            raise ValueError(f"package {package} does not exist") from None

    def get_version(self, package):
        """Return the installed version of a package."""
        pkg = self._apt_pkg(package)
        inst = pkg.installed
        if not inst:
            raise ValueError(f"package {package} does not exist")
        return inst.version

    def get_available_version(self, package):
        """Return the latest available version of a package."""
        return self._apt_pkg(package).candidate.version

    def get_dependencies(self, package):
        """Return a list of packages a package depends on."""
        cur_ver = self._apt_pkg(package).installed
        if not cur_ver:
            # happens with virtual packages
            return []
        return [
            d[0].name
            for d in cur_ver.get_dependencies("Depends", "PreDepends", "Recommends")
        ]

    def get_source(self, package):
        """Return the source package name for a package."""
        if self._apt_pkg(package).installed:
            return self._apt_pkg(package).installed.source_name
        if self._apt_pkg(package).candidate:
            return self._apt_pkg(package).candidate.source_name
        raise ValueError(f"package {package} does not exist")

    def get_package_origin(self, package):
        """Return package origin.

        Return the repository name from which a package was installed, or None
        if it cannot be determined.

        Throw ValueError if package is not installed.
        """
        pkg = self._apt_pkg(package).installed
        if not pkg:
            raise ValueError("package is not installed")
        for origin in pkg.origins:
            if origin.origin:
                return origin.origin
        return None

    def is_distro_package(self, package):
        """Check if a package is a genuine distro package.

        Return True for a native distro package, False if it comes from a
        third-party source.
        """
        pkg = self._apt_pkg(package)
        # some PPA packages have installed version None, see LP#252734
        if pkg.installed and pkg.installed.version is None:
            return False

        if not pkg.candidate:
            return False

        origins = {o.origin for o in pkg.candidate.origins}

        distro_name = self.get_os_version()[0]
        if distro_name in origins:
            return True

        # on Ubuntu system-image we might not have any /var/lib/apt/lists
        if origins == {""} and os.path.exists("/etc/system-image/channel.ini"):
            return True

        return False

    def is_native_origin_package(self, package):
        """Check if a package originated from a native location.

        Return True for a package which came from an origin which is listed in
        native-origins.d, False if it comes from a third-party source.
        """
        pkg = self._apt_pkg(package)
        # some PPA packages have installed version None, see LP#252734
        if pkg.installed and pkg.installed.version is None:
            return False

        native_origins = []
        for f in glob.glob("/etc/apport/native-origins.d/*"):
            try:
                with open(f, encoding="utf-8") as fd:
                    for line in fd:
                        line = line.strip()
                        if line:
                            native_origins.append(line)
            except OSError:
                pass

        if pkg.candidate and pkg.candidate.origins:  # might be None
            for o in pkg.candidate.origins:
                if o.origin in native_origins:
                    return True
        return False

    @staticmethod
    def get_lp_binary_package(release, package, version, arch):
        """Get Launchpad URL and SHA1 sum for the given binary package version."""
        # allow unauthenticated downloads
        apt.apt_pkg.config.set("APT::Get::AllowUnauthenticated", "True")
        # pylint: disable=import-outside-toplevel
        from launchpadlib.launchpad import Launchpad

        launchpad = Launchpad.login_anonymously(
            "apport-retrace", "production", version="devel"
        )
        ubuntu = launchpad.distributions["ubuntu"]
        series = ubuntu.getSeries(name_or_version=release.split()[-1])
        das = series.getDistroArchSeries(archtag=arch)
        primary = ubuntu.getArchive(name="primary")
        bpph = primary.getPublishedBinaries(
            binary_name=package,
            version=version,
            distro_arch_series=das,
            ordered=False,
            exact_match=True,
        )
        if not bpph:
            return (None, None)
        bf_urls = ""
        for bp in bpph:
            if bp.status == "Deleted":
                continue
            if not bp.architecture_specific:
                # include_meta is required to get the sha1
                bf_urls = bp.binaryFileUrls(include_meta=True)
                break
            if bp.distro_arch_series_link.endswith(arch):
                bf_urls = bp.binaryFileUrls(include_meta=True)
                break
        if not bf_urls:
            return (None, None)
        for bf in bf_urls:
            # return the first binary file url since there being more than one
            # is theoretical
            return (bf["url"], bf["sha1"])

    @staticmethod
    def json_request(url, entries=False):
        """Open, read and parse the JSON of a URL.

        Set entries to True when the json data returned by Launchpad
        has a dictionary with an entries key which contains the data
        desired.
        """
        try:
            with urllib.request.urlopen(url) as response:
                content = response.read()
        except (urllib.error.URLError, urllib.error.HTTPError):
            apport.logging.warning("cannot connect to: %s", urllib.parse.unquote(url))
            return None
        except OSError:
            apport.logging.warning(
                "failure reading data at: %s", urllib.parse.unquote(url)
            )
            return None
        if isinstance(content, bytes):
            content = content.decode("utf-8")
        if entries:
            return json.loads(content)["entries"]
        return json.loads(content)

    @staticmethod
    def get_lp_source_package(package, version):
        """Get files from Launchpad for the given source package version."""
        # pylint: disable=import-outside-toplevel
        from launchpadlib.launchpad import Launchpad

        launchpad = Launchpad.login_anonymously(
            "apport-retrace", "production", version="devel"
        )
        ubuntu = launchpad.distributions["ubuntu"]
        primary = ubuntu.getArchive(name="primary")
        pss = primary.getPublishedSources(
            source_name=package, version=version, exact_match=True
        )
        if not pss:
            return None
        sfus = ""
        for ps in pss:
            if ps.status == "Deleted":
                continue
            sfus = ps.sourceFileUrls()
            # use the first entry as they are sorted chronologically
            break
        if not sfus:
            return None
        source_files = []
        for sfu in sfus:
            source_files.append(sfu)
        return source_files

    def get_architecture(self, package):
        """Return the architecture of a package.

        This might differ on multiarch architectures (e. g. an i386 Firefox
        package on a x86_64 system)
        """
        if self._apt_pkg(package).installed:
            return self._apt_pkg(package).installed.architecture or "unknown"
        if self._apt_pkg(package).candidate:
            return self._apt_pkg(package).candidate.architecture or "unknown"
        raise ValueError(f"package {package} does not exist")

    def get_files(self, package):
        """Return list of files shipped by a package."""
        output = self._call_dpkg(["-L", package])
        if output is None:
            return None
        return [f for f in output.splitlines() if not f.startswith("diverted")]

    def get_modified_files(self, package):
        """Return list of all modified files of a package."""
        # get the maximum mtime of package files that we consider unmodified
        listfile = f"/var/lib/dpkg/info/{package}:{self.get_system_architecture()}.list"
        if not os.path.exists(listfile):
            listfile = f"/var/lib/dpkg/info/{package}.list"
        try:
            s = os.stat(listfile)
            if not stat.S_ISREG(s.st_mode):
                raise OSError
            max_time = max(s.st_mtime, s.st_ctime)
        except OSError:
            return []

        # create a list of files with a newer timestamp for md5sum'ing
        sums = b""
        sumfile = (
            f"/var/lib/dpkg/info/{package}:{self.get_system_architecture()}.md5sums"
        )
        if not os.path.exists(sumfile):
            sumfile = f"/var/lib/dpkg/info/{package}.md5sums"
            if not os.path.exists(sumfile):
                # some packages do not ship md5sums
                return []

        with open(sumfile, "rb") as fd:
            for line in fd:
                try:
                    # ignore lines with NUL bytes (happens, LP#96050)
                    if b"\0" in line:
                        apport.logging.warning(
                            "%s contains NUL character, ignoring line", sumfile
                        )
                        continue
                    words = line.split()
                    if not words:
                        apport.logging.warning(
                            "%s contains empty line, ignoring line", sumfile
                        )
                        continue
                    s = os.stat(f"/{words[-1].decode('UTF-8')}".encode("UTF-8"))
                    if max(s.st_mtime, s.st_ctime) <= max_time:
                        continue
                except OSError:
                    pass

                sums += line

        if sums:
            return self._check_files_md5(sums)
        return []

    def get_modified_conffiles(self, package):
        """Return modified configuration files of a package.

        Return a file name -> file contents map of all configuration files of
        package. Please note that apport.hookutils.attach_conffiles() is the
        official user-facing API for this, which will ask for confirmation and
        allows filtering.
        """
        dpkg = subprocess.run(
            ["dpkg-query", "-W", "--showformat=${Conffiles}", "--", package],
            check=False,
            stdout=subprocess.PIPE,
        )

        if dpkg.returncode != 0:
            return {}

        modified = {}
        for line in dpkg.stdout.decode().splitlines():
            if not line:
                continue
            # just take the first two fields, to not stumble over obsolete
            # conffiles
            path, default_md5sum = line.strip().split()[:2]

            if os.path.exists(path):
                try:
                    with open(path, "rb") as fd:
                        contents = fd.read()
                    m = hashlib.md5()
                    m.update(contents)
                    calculated_md5sum = m.hexdigest()

                    if calculated_md5sum != default_md5sum:
                        modified[path] = contents
                except OSError as error:
                    modified[path] = f"[inaccessible: {str(error)}]"
            else:
                modified[path] = "[deleted]"

        return modified

    @staticmethod
    def __fgrep_files(pattern, file_list):
        """Call fgrep for a pattern on given file list and return the first
        matching file, or None if no file matches."""
        match = None
        slice_size = 100
        i = 0

        while not match and i < len(file_list):
            fgrep = subprocess.run(
                ["fgrep", "-lxm", "1", "--", pattern] + file_list[i : (i + slice_size)],
                check=False,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            if fgrep.returncode == 0:
                match = fgrep.stdout.decode("UTF-8")
            i += slice_size

        return match

    def __fgrep_list_for_path(
        self, path: str, likely_lists: list[str], all_lists: list[str]
    ) -> str | None:
        # first check the likely packages
        match = self.__fgrep_files(path, likely_lists)
        if not match:
            match = self.__fgrep_files(path, all_lists)
        if not match:
            usrmerge_path = _usr_merge_alternative(path)
            if usrmerge_path:
                match = self.__fgrep_files(usrmerge_path, likely_lists)
                if not match:
                    match = self.__fgrep_files(usrmerge_path, all_lists)
        return match

    def get_file_package(
        self, file, uninstalled=False, map_cachedir=None, release=None, arch=None
    ):
        """Return the package a file belongs to.

        Return None if the file is not shipped by any package.

        If uninstalled is True, this will also find files of uninstalled
        packages; this is very expensive, though, and needs network access and
        lots of CPU and I/O resources. In this case, map_cachedir can be set to
        an existing directory which will be used to permanently store the
        downloaded maps. If it is not set, a temporary directory will be used.
        Also, release and arch can be set to a foreign release/architecture
        instead of the one from the current system.
        """
        if uninstalled:
            return self._search_contents(file, map_cachedir, release, arch)

        # check if the file is a diversion
        dpkg = subprocess.run(
            ["dpkg-divert", "--list", file],
            check=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        out = dpkg.stdout.decode("UTF-8")
        if dpkg.returncode == 0 and out:
            pkg = out.split()[-1]
            if pkg != "hardening-wrapper":
                return pkg

        fname = os.path.splitext(os.path.basename(file))[0].lower()

        all_lists = []
        likely_lists = []
        for f in glob.glob("/var/lib/dpkg/info/*.list"):
            p = os.path.splitext(os.path.basename(f))[0].lower().split(":")[0]
            if p in fname or fname in p:
                likely_lists.append(f)
            else:
                all_lists.append(f)

        match = self.__fgrep_list_for_path(file, likely_lists, all_lists)
        if match:
            return os.path.splitext(os.path.basename(match))[0].split(":")[0]

        return None

    @staticmethod
    @functools.cache
    def get_system_architecture():
        """Return the architecture of the system, in the notation used by the
        particular distribution."""
        dpkg = subprocess.run(
            ["dpkg", "--print-architecture"], check=True, stdout=subprocess.PIPE
        )
        arch = dpkg.stdout.decode().strip()
        assert arch
        return arch

    def get_library_paths(self, architecture: str | None = None) -> str:
        """Return a list of default library search paths for the architecture.

        If architecture is not specified, the system archictecture is used.

        The entries should be separated with a colon ':', like for
        $LD_LIBRARY_PATH. This needs to take any multiarch directories into
        account.
        """
        if architecture is None:
            architecture = self.get_system_architecture()
        multiarch_triplet = _ARCH_MULTIARCH_TRIPLET_MAPPING.get(architecture)
        if multiarch_triplet is None:
            raise NotImplementedError(
                f"Missing multi-arch triplet information"
                f" for architecture {architecture}"
            )
        return f"/lib/{multiarch_triplet}:/usr/lib/{multiarch_triplet}:/lib:/usr/lib"

    def set_mirror(self, url):
        """Explicitly set a distribution mirror URL for operations that need to
        fetch distribution files/packages from the network.

        By default, the mirror will be read from the system configuration
        files.
        """
        self._mirror = url

        # purge our contents dir cache
        try:
            if self._contents_dir:
                shutil.rmtree(self._contents_dir)
                self._contents_dir = None
        except AttributeError:
            pass

    def get_source_tree(self, srcpackage, output_dir, version=None, sandbox=None):
        # TODO: Split into smaller functions/methods
        # pylint: disable=too-many-branches,too-many-locals
        """Download source package and unpack it into output_dir.

        This also has to care about applying patches etc., so that output_dir
        will eventually contain the actually compiled source. output_dir needs
        to exist and should be empty.

        If version is given, this particular version will be retrieved.
        Otherwise this will fetch the latest available version.

        If sandbox is given, it calls apt-get source in that sandbox, otherwise
        it uses the system apt configuration.

        If apt_update is True, it will call apt-get update before apt-get
        source. This is mostly necessary for freshly created sandboxes.

        Return the directory that contains the actual source root directory
        (which might be a subdirectory of output_dir). Return None if the
        source is not available.
        """
        # configure apt for sandbox
        env = os.environ.copy()
        if sandbox:
            # hard to change, pylint: disable=consider-using-with
            f = tempfile.NamedTemporaryFile("w+")
            f.write(
                f'Dir "{sandbox}";\n'
                f'Dir::State::Status "/var/lib/dpkg/status";\n'
                f'Debug::NoLocking "true";\n'
            )
            f.flush()
            env["APT_CONFIG"] = f.name

        if sandbox and not glob.glob(f"{sandbox}/var/lib/apt/lists/*Sources"):
            subprocess.call(["apt-get", "-qq", "update"], env=env)

        # fetch source tree
        argv = ["apt-get", "-qq", "--assume-yes", "source", srcpackage]
        if version:
            argv[-1] += f"={version}"
        try:
            if subprocess.call(argv, cwd=output_dir, env=env) != 0:
                if not version:
                    return None
                sf_urls = self.get_lp_source_package(srcpackage, version)
                if sf_urls:
                    fetchProgress = apt.progress.base.AcquireProgress()
                    fetcher = apt.apt_pkg.Acquire(fetchProgress)
                    af_queue = []
                    for sf in sf_urls:
                        af_queue.append(
                            apt.apt_pkg.AcquireFile(fetcher, sf, destdir=output_dir)
                        )
                    result = fetcher.run()
                    if result != fetcher.RESULT_CONTINUE:
                        return None
                    for dsc in glob.glob(os.path.join(output_dir, "*.dsc")):
                        subprocess.call(
                            ["dpkg-source", "-sn", "-x", dsc],
                            stdout=subprocess.PIPE,
                            cwd=output_dir,
                        )
                else:
                    return None
        except OSError:
            return None

        # find top level directory
        root = None
        for d in glob.glob(os.path.join(output_dir, f"{srcpackage}-*")):
            if os.path.isdir(d):
                root = d
        assert root, "could not determine source tree root directory"

        # apply patches on a best-effort basis
        try:
            subprocess.call(
                "(debian/rules patch || debian/rules apply-patches "
                "|| debian/rules apply-dpatches || "
                "debian/rules unpack || debian/rules patch-stamp || "
                "debian/rules setup) >/dev/null 2>&1",
                shell=True,
                cwd=root,
            )
        except OSError:
            pass

        return root

    def get_kernel_package(self):
        """Return the actual Linux kernel package name.

        This is used when the user reports a bug against the "linux" package.
        """
        # TODO: Ubuntu specific
        return f"linux-image-{os.uname()[2]}"

    def _apt_cache_root_dir(
        self, architecture: str, cache_dir: pathlib.Path | str, release: str
    ) -> str:
        """Determine APT cache root directory.

        release is the value of the report's 'DistroRelease' field or
        "system" in case the current system configuration should be used.
        """
        if architecture != self.get_system_architecture():
            aptroot_arch = architecture
        else:
            aptroot_arch = ""
        return os.path.join(cache_dir, release, aptroot_arch, "apt")

    def install_packages(
        self,
        rootdir,
        configdir,
        release,
        packages,
        verbose=False,
        cache_dir=None,
        permanent_rootdir=False,
        architecture=None,
        origins=None,
        install_dbg=True,
        install_deps=False,
    ):  # pylint: disable=too-many-arguments
        # TODO: Split into smaller functions/methods
        # pylint: disable=too-many-branches,too-many-locals
        # pylint: disable=too-many-nested-blocks,too-many-statements
        """Install packages into a sandbox (for apport-retrace).

        In order to work without any special permissions and without touching
        the running system, this should only download and unpack packages into
        the given root directory, not install them into the system.

        configdir points to a directory with by-release configuration files for
        the packaging system; this is completely dependent on the backend
        implementation, the only assumption is that this looks into
        configdir/release/, so that you can use retracing for multiple
        DistroReleases. As a special case, if configdir is None, it uses the
        current system configuration, and "release" is ignored.

        release is the value of the report's 'DistroRelease' field.

        packages is a list of ('packagename', 'version') tuples. If the version
        is None, it should install the most current available version.

        If cache_dir is given, then the downloaded packages will be stored
        there, to speed up subsequent retraces.

        If permanent_rootdir is True, then the sandbox created from the
        downloaded packages will be reused, to speed up subsequent retraces.

        If architecture is given, the sandbox will be created with packages of
        the given architecture (as specified in a report's "Architecture"
        field). If not given it defaults to the host system's architecture.

        If origins is given, the sandbox will be created with apt data sources
        for foreign origins.

        If install_deps is True, then the dependencies of packages will also
        be installed.

        Return a string with outdated packages, or an empty string if all
        packages were installed.

        If something is wrong with the environment (invalid configuration,
        package servers down, etc.), this should raise a SystemError with a
        meaningful error message.
        """
        if not architecture:
            architecture = self.get_system_architecture()
        if not configdir:
            apt_dir = "/etc/apt"
            self._current_release_codename = self.get_distro_codename()
        else:
            # support architecture specific config, fall back to global config
            apt_dir = os.path.join(configdir, release)
            if architecture != self.get_system_architecture():
                arch_apt_dir = os.path.join(configdir, release, architecture)
                arch_old_sources = os.path.join(arch_apt_dir, "sources.list")
                arch_sources_dir = os.path.join(arch_apt_dir, "sources.list.d")
                if os.path.exists(arch_old_sources) or (
                    os.path.exists(arch_sources_dir) and os.listdir(arch_sources_dir)
                ):
                    apt_dir = arch_apt_dir

            # set mirror for get_file_package()
            try:
                self.set_mirror(self._get_primary_mirror_from_apt_sources(apt_dir))
            except SystemError as error:
                apport.logging.warning("cannot determine mirror: %s", str(error))

            # set current release code name for _distro_release_to_codename
            with open(
                os.path.join(configdir, release, "codename"), encoding="utf-8"
            ) as f:
                self._current_release_codename = f.read().strip()

        # create apt sandbox
        if cache_dir:
            tmp_aptroot = False
            aptroot = self._apt_cache_root_dir(
                architecture, cache_dir, release if configdir else "system"
            )
            if not os.path.isdir(aptroot):
                os.makedirs(aptroot)
        else:
            tmp_aptroot = True
            aptroot = tempfile.mkdtemp()

        apt.apt_pkg.config.set("APT::Architecture", architecture)
        apt.apt_pkg.config.set("Acquire::Languages", "none")

        if verbose:
            fetchProgress = apt.progress.text.AcquireProgress()
        else:
            fetchProgress = apt.progress.base.AcquireProgress()
        if not tmp_aptroot:
            apt_cache = self._sandbox_cache(
                aptroot,
                apt_dir,
                fetchProgress,
                self.get_distro_name(),
                self._current_release_codename,
                origins,
                architecture,
            )
        else:
            self._build_apt_sandbox(
                aptroot,
                apt_dir,
                self.get_distro_name(),
                self._current_release_codename,
                origins,
            )
            apt_cache = apt.Cache(rootdir=os.path.abspath(aptroot))
            try:
                apt_cache.update(fetchProgress)
            except apt.cache.FetchFailedException as error:
                raise SystemError(str(error)) from error
            apt_cache.open()

        archivedir = apt.apt_pkg.config.find_dir("Dir::Cache::archives")

        obsolete = ""

        src_records = apt.apt_pkg.SourceRecords()

        # read original package list
        pkg_list = os.path.join(rootdir, "packages.txt")
        pkg_versions = {}
        if os.path.exists(pkg_list):
            with open(pkg_list, encoding="utf-8") as f:
                for line in f:
                    line = line.strip()
                    if not line:
                        continue
                    (p, v) = line.split()
                    pkg_versions[p] = v

        # mark packages for installation
        real_pkgs = set()
        lp_cache = {}
        fetcher = apt.apt_pkg.Acquire(fetchProgress)
        # need to keep AcquireFile references
        acquire_queue = []
        # add any dependencies to the packages list
        if install_deps:
            deps = []
            for pkg, ver in packages:
                try:
                    cache_pkg = apt_cache[pkg]
                except KeyError:
                    m = f"package {pkg.replace('%', '%%')} does not exist, ignoring"
                    obsolete += f"{m}\n"
                    apport.logging.warning("%s", m)
                    continue
                for dep in cache_pkg.candidate.dependencies:
                    # the dependency may be satisfied by a different package
                    if dep[0].name not in apt_cache:
                        dep[0] = apt_cache.get_providing_packages(dep[0].name)[0]
                    # the version in dep is the one from pkg's dependencies,
                    # so use the version from the cache
                    dep_pkg_vers = apt_cache[dep[0].name].candidate.version
                    # if the dependency is in the list of packages we don't
                    # need to look up its dependencies again
                    if dep[0].name in [pkg[0] for pkg in packages]:
                        continue
                    # if the package is already extracted in the sandbox
                    # because the report needs that package we don't want to
                    # install a newer version which may cause a CRC mismatch
                    # with the installed dbg symbols
                    if dep[0].name in pkg_versions:
                        inst_version = pkg_versions[dep[0].name]
                        if self.compare_versions(inst_version, dep_pkg_vers) > -1:
                            deps.append((dep[0].name, inst_version))
                        else:
                            deps.append((dep[0].name, dep_pkg_vers))
                    else:
                        deps.append((dep[0].name, dep_pkg_vers))
                    if dep[0].name not in [pkg[0] for pkg in packages]:
                        packages.append((dep[0].name, None))
            packages.extend(deps)

        for pkg, ver in packages:
            try:
                cache_pkg = apt_cache[pkg]
            except KeyError:
                m = f"package {pkg.replace('%', '%%')} does not exist, ignoring"
                obsolete += f"{m}\n"
                apport.logging.warning("%s", m)
                continue

            # try to select matching version
            try:
                if ver:
                    cache_pkg.candidate = cache_pkg.versions[ver]
            except KeyError:
                (lp_url, sha1sum) = self.get_lp_binary_package(
                    release, pkg, ver, architecture
                )
                if lp_url:
                    acquire_queue.append(
                        apt.apt_pkg.AcquireFile(
                            fetcher, lp_url, hash=f"sha1:{sha1sum}", destdir=archivedir
                        )
                    )
                    lp_cache[pkg] = ver
                else:
                    obsolete += (
                        f"{pkg} version {ver} required,"
                        f" but {cache_pkg.candidate.version} is available\n"
                    )
                    ver = cache_pkg.candidate.version

            candidate = cache_pkg.candidate
            real_pkgs.add(pkg)

            if permanent_rootdir:
                virtual_mapping = self._virtual_mapping(aptroot)
                # Remember all the virtual packages that this package provides,
                # so that if we encounter that virtual package as a
                # Conflicts/Replaces later, we know to remove this package from
                # the cache.
                for p in candidate.provides:
                    virtual_mapping.setdefault(p, set()).add(pkg)
                conflicts = []
                if "Conflicts" in candidate.record:
                    conflicts += apt.apt_pkg.parse_depends(
                        candidate.record["Conflicts"]
                    )
                if "Replaces" in candidate.record:
                    conflicts += apt.apt_pkg.parse_depends(candidate.record["Replaces"])
                for conflict in conflicts:
                    # if the package conflicts with itself its wonky e.g.
                    # gdb in artful
                    if conflict[0][0] == candidate.package.name:
                        continue
                    # apt_pkg.parse_depends needs to handle the or operator,
                    # but as policy states it is invalid to use that in
                    # Replaces/Depends, we can safely choose the first value
                    # here.
                    conflict = conflict[0]
                    if apt_cache.is_virtual_package(conflict[0]):
                        try:
                            providers = virtual_mapping[conflict[0]]
                        except KeyError:
                            # We may not have seen the virtual package that
                            # this conflicts with, so we can assume it's not
                            # unpacked into the sandbox.
                            continue
                        for p in providers:
                            # if the candidate package being installed
                            # conflicts with but also provides a virtual
                            # package don't act on the candidate e.g.
                            # libpam-modules and libpam-mkhomedir in artful
                            if p == candidate.package.name:
                                continue
                            debs = os.path.join(archivedir, f"{p}_*.deb")
                            for path in glob.glob(debs):
                                ver = self._deb_version(path)
                                if apt.apt_pkg.check_dep(ver, conflict[2], conflict[1]):
                                    os.unlink(path)
                            try:
                                del pkg_versions[p]
                            except KeyError:
                                pass
                        del providers
                    else:
                        debs = os.path.join(archivedir, f"{conflict[0]}_*.deb")
                        for path in glob.glob(debs):
                            ver = self._deb_version(path)
                            if apt.apt_pkg.check_dep(ver, conflict[2], conflict[1]):
                                os.unlink(path)
                                try:
                                    del pkg_versions[conflict[0]]
                                except KeyError:
                                    pass

            if candidate.architecture != "all" and install_dbg:
                try:
                    dbg_pkg = f"{pkg}-dbg"
                    dbg = apt_cache[dbg_pkg]
                    pkg_found = False
                    # try to get the same version as pkg
                    if ver:
                        try:
                            dbg.candidate = dbg.versions[ver]
                            pkg_found = True
                        except KeyError:
                            (lp_url, sha1sum) = self.get_lp_binary_package(
                                release, dbg_pkg, ver, architecture
                            )
                            if lp_url:
                                acquire_queue.append(
                                    apt.apt_pkg.AcquireFile(
                                        fetcher,
                                        lp_url,
                                        hash=f"sha1:{sha1sum}",
                                        destdir=archivedir,
                                    )
                                )
                                lp_cache[dbg_pkg] = ver
                                pkg_found = True
                            # if it can't be found in Launchpad failover to a
                            # code path that'll use -dbgsym packages
                            else:
                                raise
                    if not pkg_found:
                        try:
                            dbg.candidate = dbg.versions[candidate.version]
                        except KeyError:
                            obsolete += (
                                f"outdated -dbg package for {pkg}:"
                                f" package version {ver}"
                                f" -dbg version {dbg.candidate.version}\n"
                            )
                    real_pkgs.add(dbg_pkg)
                except KeyError:
                    # install all -dbg from the source package; lookup() just
                    # works from the current list pointer, we always need to
                    # start from the beginning
                    src_records.restart()
                    if src_records.lookup(candidate.source_name):
                        # ignore transitional packages
                        # False positive, see
                        # https://github.com/PyCQA/pylint/issues/7122
                        # pylint: disable=not-an-iterable
                        dbgs = [
                            p
                            for p in src_records.binaries
                            if p.endswith("-dbg")
                            and p in apt_cache
                            and "transitional" not in apt_cache[p].candidate.description
                        ]
                        # if a specific version of a package was requested
                        # only install dbg pkgs whose version matches
                        if ver:
                            for dbg in dbgs:
                                if apt_cache[dbg].candidate.version != ver:
                                    dbgs.remove(dbg)
                    else:
                        dbgs = []
                    if dbgs:
                        for p in dbgs:
                            # if the package has already been added to
                            # real_pkgs don't search for it again
                            if p in real_pkgs:
                                continue
                            pkg_found = False
                            # prefer the version requested
                            if ver:
                                try:
                                    apt_cache[p].candidate = apt_cache[p].versions[ver]
                                    pkg_found = True
                                except KeyError:
                                    (lp_url, sha1sum) = self.get_lp_binary_package(
                                        release, p, ver, architecture
                                    )
                                    if lp_url:
                                        acquire_queue.append(
                                            apt.apt_pkg.AcquireFile(
                                                fetcher,
                                                lp_url,
                                                hash=f"sha1:{sha1sum}",
                                                destdir=archivedir,
                                            )
                                        )
                                        lp_cache[p] = ver
                                        pkg_found = True
                            if not pkg_found:
                                try:
                                    apt_cache[p].candidate = apt_cache[p].versions[
                                        candidate.version
                                    ]
                                except KeyError:
                                    # we don't really expect that, but it's
                                    # possible that other binaries have a
                                    # different version
                                    pass
                            real_pkgs.add(p)
                    else:
                        pkg_found = False
                        dbgsym_pkg = f"{pkg}-dbgsym"
                        try:
                            dbgsym = apt_cache[dbgsym_pkg]
                            real_pkgs.add(dbgsym_pkg)
                            # prefer the version requested
                            if ver:
                                try:
                                    dbgsym.candidate = dbgsym.versions[ver]
                                    pkg_found = True
                                except KeyError:
                                    (lp_url, sha1sum) = self.get_lp_binary_package(
                                        release, dbgsym_pkg, ver, architecture
                                    )
                                    if lp_url:
                                        acquire_queue.append(
                                            apt.apt_pkg.AcquireFile(
                                                fetcher,
                                                lp_url,
                                                hash=f"sha1:{sha1sum}",
                                                destdir=archivedir,
                                            )
                                        )
                                        lp_cache[dbgsym_pkg] = ver
                                        pkg_found = True
                            if not pkg_found:
                                try:
                                    dbgsym.candidate = dbgsym.versions[
                                        candidate.version
                                    ]
                                except KeyError:
                                    obsolete += (
                                        f"outdated debug symbol package"
                                        f" for {pkg}: package version"
                                        f" {candidate.version}"
                                        f" dbgsym version"
                                        f" {dbgsym.candidate.version}\n"
                                    )

                        except KeyError:
                            if ver:
                                (lp_url, sha1sum) = self.get_lp_binary_package(
                                    release, dbgsym_pkg, ver, architecture
                                )
                                if lp_url:
                                    acquire_queue.append(
                                        apt.apt_pkg.AcquireFile(
                                            fetcher,
                                            lp_url,
                                            hash=f"sha1:{sha1sum}",
                                            destdir=archivedir,
                                        )
                                    )
                                    lp_cache[dbgsym_pkg] = ver
                                    pkg_found = True
                            if not pkg_found:
                                obsolete += f"no debug symbol package found for {pkg}\n"

        # unpack packages, weed out the ones that are already installed (for
        # permanent sandboxes)
        logger = logging.getLogger(__name__)
        requested_pkgs = dict(packages)
        for p in real_pkgs.copy():
            if p in requested_pkgs:
                if requested_pkgs[p] is None:
                    # We already have the latest version of this package
                    if pkg_versions.get(p) == apt_cache[p].candidate.version:
                        logger.debug(
                            "Removing %s which is already the right version", p
                        )
                        real_pkgs.remove(p)
                    else:
                        logger.debug(
                            "Installing %s version %s",
                            p,
                            apt_cache[p].candidate.version,
                        )
                        apt_cache[p].mark_install(False, False)
                elif pkg_versions.get(p) != requested_pkgs[p]:
                    logger.debug(
                        "Installing %s version %s", p, apt_cache[p].candidate.version
                    )
                    apt_cache[p].mark_install(False, False)
                elif pkg_versions.get(p) != apt_cache[p].candidate.version:
                    logger.debug(
                        "Installing %s version %s", p, apt_cache[p].candidate.version
                    )
                    apt_cache[p].mark_install(False, False)
                else:
                    logger.debug("Removing %s which is already the right version", p)
                    real_pkgs.remove(p)
            elif pkg_versions.get(p) != apt_cache[p].candidate.version:
                logger.debug("Installing %s", p)
                apt_cache[p].mark_install(False, False)
            else:
                logger.debug("Removing %s which is already the right version", p)
                real_pkgs.remove(p)

        last_written = time.time()
        # fetch packages
        try:
            apt_cache.fetch_archives(fetcher=fetcher)
        except apt.cache.FetchFailedException as error:
            apport.logging.error(
                "Package download error, try again later: %s", str(error)
            )
            sys.exit(1)  # transient error

        if verbose:
            print("Extracting downloaded debs...")
        # False positive, see https://github.com/PyCQA/pylint/issues/7122
        for i in fetcher.items:  # pylint: disable=not-an-iterable
            if not i.destfile.endswith("deb"):
                continue
            out = subprocess.check_output(["dpkg-deb", "--show", i.destfile]).decode()
            (p, v) = out.strip().split()
            if (
                not permanent_rootdir
                or p not in pkg_versions
                or os.path.getctime(i.destfile) > last_written
            ):
                # don't extract the same version of the package if it is
                # already extracted
                if pkg_versions.get(p) == v:
                    pass
                # don't extract the package if it is a different version than
                # the one we want to extract from Launchpad
                elif p in lp_cache and lp_cache[p] != v:
                    pass
                else:
                    subprocess.check_call(["dpkg", "-x", i.destfile, rootdir])
                    pkg_versions[p] = v
            pkg_name = os.path.basename(i.destfile).split("_", 1)[0]
            # because a package may exist multiple times in the fetcher it may
            # have already been removed
            if pkg_name in real_pkgs:
                real_pkgs.remove(pkg_name)

        # update package list
        pkgs = list(pkg_versions.keys())
        pkgs.sort()
        with open(pkg_list, "w", encoding="utf-8") as f:
            for p in pkgs:
                f.write(p)
                f.write(" ")
                f.write(pkg_versions[p])
                f.write("\n")

        if tmp_aptroot:
            shutil.rmtree(aptroot)

        # check bookkeeping that apt fetcher really got everything
        assert (
            not real_pkgs
        ), f"apt fetcher did not fetch these packages: {' '.join(real_pkgs)}"

        if permanent_rootdir:
            self._save_virtual_mapping(aptroot)

        return obsolete

    def package_name_glob(self, nameglob):
        """Return known package names which match given glob."""
        return glob.fnmatch.filter(self._cache().keys(), nameglob)

    #
    # Internal helper methods
    #

    @staticmethod
    def _call_dpkg(args):
        """Call dpkg with given arguments and return output, or return None on
        error."""
        dpkg = subprocess.run(
            ["dpkg"] + args, check=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        if dpkg.returncode == 0:
            return dpkg.stdout.decode("UTF-8")
        raise ValueError("package does not exist")

    @staticmethod
    def _check_files_md5(sumfile):
        """Call md5sum.

        This is separate from get_modified_files so that it is automatically
        testable.
        """
        if os.path.exists(sumfile):
            args = [sumfile]
            stdin = None
        else:
            assert isinstance(sumfile, bytes), "md5sum list value must be a byte array"
            args = []
            stdin = sumfile
        md5sum = subprocess.run(
            ["/usr/bin/md5sum", "-c"] + args,
            check=False,
            input=stdin,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd="/",
            env={},
        )

        # if md5sum succeeded, don't bother parsing the output
        if md5sum.returncode == 0:
            return []
        out = md5sum.stdout.decode("UTF-8", errors="replace")

        mismatches = []
        for line in out.splitlines():
            if line.endswith("FAILED"):
                mismatches.append(line.rsplit(":", 1)[0])

        return mismatches

    @staticmethod
    def _get_primary_mirror_from_apt_sources(apt_dir: str) -> str:
        """Heuristically determine primary mirror from an apt sources.list."""
        uri = None
        sources = _load_all_sources(apt_dir)
        for source in sources:
            if source.disabled or source.invalid:
                continue
            # TODO: remove the typeched branch once python-apt commit
            # bdcb2550cb6f623f4556bf6581a040642f29dd28 is available
            # in a proper release
            if isinstance(source, SourceEntry):
                if source.type == "deb":
                    uri = source.uri or ""
            elif "deb" in source.types:
                uri = source.uris[0]

            if uri is not None:
                if uri.startswith("mirror+file:"):
                    mirrors = _read_mirror_file(uri)
                    assert mirrors
                    return mirrors[0]
                return uri

        raise SystemError(
            "cannot determine default mirror:"
            f" couldn't find configured source contains the `deb` type in {apt_dir}"
        )

    def _get_mirror(self, arch: str) -> str:
        """Return the distribution mirror URL.

        If it has not been set yet, it will be read from the system
        configuration.
        """
        if not self._mirror:
            self._mirror = self._get_primary_mirror_from_apt_sources("/etc/apt")
        return _map_mirror_to_arch(self._mirror, arch)

    def _ppa_archive_url(self, user: str, distro: str, ppa_name: str) -> str:
        return f"{self._launchpad_base}/~{user}/+archive/{distro}/{ppa_name}"

    def _distro_release_to_codename(self, release):
        """Map a DistroRelease: field value to a release code name."""
        # if we called install_packages() with a configdir, we can read the
        # codename from there
        if self._current_release_codename is not None:
            return self._current_release_codename

        raise NotImplementedError(
            f"Cannot map DistroRelease '{release}' to a code name"
            " without install_packages()"
        )

    def _fetch_contents_file(
        self, contents_filename: str, mtime: float | None, dist: str, arch: str
    ) -> bool:
        """Returns True on success (including when no update is needed)."""
        url = f"{self._get_mirror(arch)}/dists/{dist}/Contents-{arch}.gz"
        if mtime:
            req = urllib.request.Request(url, method="HEAD")
            with contextlib.suppress(OSError):
                with urllib.request.urlopen(req) as res:
                    modified_str = res.getheader("last-modified", None)
                    if modified_str:
                        modified = datetime.datetime.strptime(
                            modified_str, "%a, %d %b %Y %H:%M:%S %Z"
                        )
                        if modified <= datetime.datetime.fromtimestamp(mtime):
                            return True
                # don't update the file if it is empty
                if res.getheader("content-length", None) == "40":
                    return True

        self._contents_update = True
        try:
            # hard to change, pylint: disable=consider-using-with
            src = urllib.request.urlopen(url)
        except OSError:
            # we ignore non-existing pockets, but we do crash
            # if the release pocket doesn't exist
            if "-" not in dist:
                raise
            return False

        with open(contents_filename, "wb") as f:
            while True:
                data = src.read(1000000)
                if not data:
                    break
                f.write(data)
        src.close()
        assert os.path.exists(contents_filename)
        return True

    def _get_contents_file(self, map_cachedir: str, dist: str, arch: str) -> str | None:
        contents_filename = os.path.join(map_cachedir, f"{dist}-Contents-{arch}.gz")
        # check if map exists and is younger than a day; if not, we need
        # to refresh it
        try:
            mtime = os.stat(contents_filename).st_mtime
            age = int(time.time() - mtime)
        except OSError:
            mtime = None
            age = None

        if age is None or age >= 86400:
            if not self._fetch_contents_file(contents_filename, mtime, dist, arch):
                return None

        return contents_filename

    @staticmethod
    def _update_given_file2pkg_mapping(
        file2pkg: dict[bytes, bytes], contents_filename: str, dist: str
    ) -> None:
        path_exclude_pattern = re.compile(_EXCLUDED_PATHS.encode())
        with gzip.open(contents_filename, "rb") as contents:
            line_num = 0
            for line in contents:
                line_num += 1
                # the first 32 lines are descriptive only for these
                # releases
                if dist in {"trusty", "xenial"} and line_num < 33:
                    continue

                if path_exclude_pattern.match(line):
                    continue
                parts = line.split()
                path = parts[0]
                package = parts[-1].split(b",")[0].split(b"/")[-1]
                if path in file2pkg:
                    if package == file2pkg[path]:
                        continue
                    # if the package was updated use the update
                    # b/c everyone should have packages from
                    # -updates and -security installed
                file2pkg[path] = package

    def _get_file2pkg_mapping(
        self, map_cachedir: str, release: str, arch: str
    ) -> dict[bytes, bytes]:
        # this is ordered by likelihood of installation with the most common
        # last
        # XXX - maybe we shouldn't check -security and -updates if it is the
        # devel release as they will be old and empty
        for pocket in ("-proposed", "", "-security", "-updates"):
            dist = f"{release}{pocket}"
            contents_filename = self._get_contents_file(map_cachedir, dist, arch)
            if contents_filename is None:
                continue
            file2pkg = self._contents_mapping(map_cachedir, release, arch)
            # if the mapping is empty build it
            if not file2pkg or len(file2pkg) == 2:
                self._contents_update = True
            # if any of the Contents files were updated we need to update the
            # map because the ordering in which is created is important
            if self._contents_update:
                self._update_given_file2pkg_mapping(file2pkg, contents_filename, dist)
        return file2pkg

    def _search_contents(self, file, map_cachedir, release, arch):
        """Search file in Contents.gz."""
        if not map_cachedir:
            if not self._contents_dir:
                self._contents_dir = tempfile.mkdtemp()
            map_cachedir = self._contents_dir

        if arch is None:
            arch = self.get_system_architecture()
        if release is None:
            release = self.get_distro_codename()
        else:
            release = self._distro_release_to_codename(release)

        contents_mapping = self._get_file2pkg_mapping(map_cachedir, release, arch)
        # the file only needs to be saved after an update
        if self._contents_update:
            self._save_contents_mapping(map_cachedir, release, arch)
            # the update of the mapping only needs to be done once
            self._contents_update = False

        if isinstance(file, bytes):
            file = file.decode()
        if file[0] != "/":
            file = f"/{file}"
        files = [file[1:].encode()]
        usrmerge_file = _usr_merge_alternative(file)
        if usrmerge_file:
            files.append(usrmerge_file[1:].encode())
        for filename in files:
            try:
                pkg = contents_mapping[filename].decode()
                return pkg
            except KeyError:
                pass
        return None

    def _analyze_ppa_origin_string(
        self, origin: str, distro: str
    ) -> tuple[str, str] | None:
        if not origin.startswith("LP-PPA-"):
            return None

        components = origin.split("-")[2:]
        # If the PPA is unnamed, it will not appear in origin information
        # but is named ppa in Launchpad.
        try_ppa = True
        if len(components) == 1:
            components.append("ppa")
            try_ppa = False

        index = 1
        while index < len(components):
            # For an origin we can't tell where the user name ends and the
            # PPA name starts, so split on each "-" until we find a PPA
            # that exists.
            user = str.join("-", components[:index])
            ppa_name = str.join("-", components[index:])
            try:
                with contextlib.closing(
                    urllib.request.urlopen(
                        self._ppa_archive_url(
                            user=user, distro=distro, ppa_name=ppa_name
                        )
                    )
                ) as response:
                    response.read()
                    return user, ppa_name
            except (urllib.error.URLError, urllib.error.HTTPError):
                index += 1
                if index == len(components):
                    if try_ppa:
                        components.append("ppa")
                        try_ppa = False
                        index = 2
        return None

    # pylint: disable-next=no-member
    def create_ppa_source_from_origin(
        self, origin: str, distro: str, release_codename: str
    ) -> list[apt_sl.Deb822SourceEntry | SourceEntry] | None:
        """For an origin from a Launchpad PPA create sources.list content.

        distro is the distribution for which content is being created e.g.
        ubuntu.

        release_codename is the codename of the release for which content is
        being created e.g. trusty.

        Returns None if the origin is not a Launchpad PPA, a list of source
        entry objects otherwise."""
        ppa_data = self._analyze_ppa_origin_string(origin, distro)
        if not ppa_data:
            return None
        user, ppa_name = ppa_data
        debug_url = (
            f"http://ppa.launchpad.net/{user}/{ppa_name}/{distro}"
            f"/dists/{release_codename}/main/debug"
        )

        if WITH_DEB822_SUPPORT:
            main_entry = apt_sl.Deb822SourceEntry(None, "")
            main_entry.uris = [f"http://ppa.launchpad.net/{user}/{ppa_name}/{distro}"]
            main_entry.types = ["deb", "deb-src"]
            main_entry.comps = ["main"]
            main_entry.suites = [release_codename]

            try:
                with contextlib.closing(urllib.request.urlopen(debug_url)) as response:
                    response.read()
                debug_entry = apt_sl.Deb822SourceEntry(None, "")
                debug_entry.uris = [
                    f"http://ppa.launchpad.net/{user}/{ppa_name}/{distro}"
                ]
                debug_entry.types = ["deb"]
                debug_entry.comps = ["main/debug"]
                debug_entry.suites = [release_codename]
                return [main_entry, debug_entry]
            except (urllib.error.URLError, urllib.error.HTTPError):
                return [main_entry]
        else:
            ppa_line = (
                f"deb http://ppa.launchpad.net/{user}/{ppa_name}/{distro}"
                f" {release_codename} main"
            )
            try:
                with contextlib.closing(urllib.request.urlopen(debug_url)) as response:
                    response.read()
                add_debug = " main/debug"
            except (urllib.error.URLError, urllib.error.HTTPError):
                add_debug = ""
            return [
                SourceEntry(ppa_line + add_debug),
                SourceEntry(f"deb-src {ppa_line[4:]}"),
            ]

    @staticmethod
    def _find_source_file_from_origin(origin: str, src_list_d: str) -> str | None:
        """Find a possible source file matching the Origin field of a
        given package (i.e. a PPA)"""
        if os.path.isdir(src_list_d):
            # check to see if there is a sources.list file for the
            # origin, if there isn't try using a sources.list file
            # w/o LP-PPA-
            candidates = [os.path.join(src_list_d, f"{origin}.list")]
            if WITH_DEB822_SUPPORT:
                candidates[0:0] = [os.path.join(src_list_d, f"{origin}.source")]
            if "LP-PPA" in origin:
                stripped = origin.strip("LP-PPA-")
                if WITH_DEB822_SUPPORT:
                    candidates.append(os.path.join(f"{stripped}.source"))
                candidates.append(os.path.join(f"{stripped}.list"))

            for path in candidates:
                if os.path.exists(path):
                    return path
        return None

    def _build_apt_sandbox(
        self, apt_root, apt_dir, distro_name, release_codename, origins
    ):
        # TODO: Split into smaller functions/methods
        # pylint: disable=too-many-branches,too-many-locals,too-many-statements

        # pre-create directories, to avoid apt.Cache() printing "creating..."
        # messages on stdout
        if not os.path.exists(os.path.join(apt_root, "var", "lib", "apt")):
            os.makedirs(os.path.join(apt_root, "var", "lib", "apt", "lists", "partial"))
            os.makedirs(
                os.path.join(apt_root, "var", "cache", "apt", "archives", "partial")
            )
            os.makedirs(os.path.join(apt_root, "var", "lib", "dpkg"))
            os.makedirs(os.path.join(apt_root, "etc", "apt", "apt.conf.d"))
            os.makedirs(os.path.join(apt_root, "etc", "apt", "preferences.d"))

        # install apt sources
        src_list_d = os.path.join(apt_dir, "sources.list.d")
        dst_list_d = os.path.join(apt_root, "etc", "apt", "sources.list.d")
        if os.path.exists(dst_list_d):
            shutil.rmtree(dst_list_d)
        if os.path.isdir(src_list_d):
            shutil.copytree(src_list_d, dst_list_d)
        else:
            os.makedirs(dst_list_d)
        old_sources = os.path.join(apt_dir, "sources.list")
        if os.path.exists(old_sources):
            with open(old_sources, encoding="utf-8") as src:
                with open(
                    os.path.join(apt_root, "etc", "apt", "sources.list"),
                    "w",
                    encoding="utf-8",
                ) as dest:
                    dest.write(src.read())

        # install apt keyrings; prefer the ones from the config dir, fall back
        # to system
        trusted_gpg = os.path.join(apt_dir, "trusted.gpg")
        if os.path.exists(trusted_gpg):
            shutil.copy(trusted_gpg, os.path.join(apt_root, "etc", "apt"))
        elif os.path.exists("/etc/apt/trusted.gpg"):
            shutil.copy("/etc/apt/trusted.gpg", os.path.join(apt_root, "etc", "apt"))

        trusted_d = os.path.join(apt_root, "etc", "apt", "trusted.gpg.d")
        if os.path.exists(trusted_d):
            shutil.rmtree(trusted_d)

        if os.path.exists(f"{trusted_gpg}.d"):
            shutil.copytree(f"{trusted_gpg}.d", trusted_d)
        elif os.path.exists("/etc/apt/trusted.gpg.d"):
            shutil.copytree("/etc/apt/trusted.gpg.d", trusted_d)
        else:
            os.makedirs(trusted_d)

        if origins:
            # map an origin to a Launchpad username and PPA name
            origin_data = {}
            for origin in origins:
                # apport's report format uses unknown for packages w/o
                # an origin
                if origin == "unknown":
                    continue
                origin_path = self._find_source_file_from_origin(origin, src_list_d)
                extension = ".sources"
                if origin_path:
                    if origin_path.endswith(".list"):
                        extension = ".list"
                        with open(origin_path, encoding="utf-8") as src:
                            source_list_content = [SourceEntry(line) for line in src]
                    else:
                        source_list_content = _parse_deb822_sources(origin_path)
                else:
                    source_list_content = self.create_ppa_source_from_origin(
                        origin, distro_name, release_codename
                    )
                    if not WITH_DEB822_SUPPORT:
                        extension = ".list"
                if source_list_content:
                    with open(
                        os.path.join(
                            apt_root, "etc", "apt", "sources.list.d", origin + extension
                        ),
                        "a",
                        encoding="utf-8",
                    ) as dest:
                        dest.write(
                            "\n".join([f"{entry}\n" for entry in source_list_content])
                        )
                    for entry in source_list_content:
                        if not entry.uri or "ppa.launchpad.net" not in entry.uri:
                            continue
                        user = entry.uri.split("/")[3]
                        ppa = entry.uri.split("/")[4]
                        origin_data[origin] = (user, ppa)
                else:
                    apport.logging.warning(
                        "Could not find or create source config for %s", origin
                    )

            # install apt keyrings for PPAs
            for origin, (ppa_user, ppa_name) in origin_data.items():
                ppa_archive_url = self._ppa_archive_url(
                    user=urllib.parse.quote(ppa_user),
                    distro=distro_name,
                    ppa_name=urllib.parse.quote(ppa_name),
                )
                ppa_key = self.json_request(
                    f"{ppa_archive_url}?ws.op=getSigningKeyData"
                )
                if not ppa_key:
                    continue
                key_file = pathlib.Path(trusted_d) / f"{origin}.asc"
                key_file.write_text(ppa_key, encoding="utf-8")

    @staticmethod
    def _deb_version(pkg):
        """Return the version of a .deb file."""
        dpkg = subprocess.run(
            ["dpkg-deb", "-f", pkg, "Version"], check=True, stdout=subprocess.PIPE
        )
        out = dpkg.stdout.decode("UTF-8").strip()
        assert out
        return out

    def compare_versions(self, ver1, ver2):
        """Compare two package versions.

        Return -1 for ver < ver2, 0 for ver1 == ver2, and 1 for ver1 > ver2.
        """
        return apt.apt_pkg.version_compare(ver1, ver2)

    _distro_codename = None

    def get_distro_codename(self):
        """Get "lsb_release -sc", cache the result."""
        if self._distro_codename is None:
            try:
                info = platform.freedesktop_os_release()
                self._distro_codename = info["VERSION_CODENAME"]
            except (KeyError, OSError):
                # Fall back to query lsb_release
                lsb_release = subprocess.run(
                    ["lsb_release", "-sc"],
                    check=True,
                    stdout=subprocess.PIPE,
                    text=True,
                )
                self._distro_codename = lsb_release.stdout.strip()

        return self._distro_codename

    _distro_name = None

    def get_distro_name(self):
        """Get osname from /etc/os-release, or if that doesn't exist,
        'lsb_release -sir' output and cache the result."""
        if self._distro_name is None:
            self._distro_name = self.get_os_version()[0].lower()
            if " " in self._distro_name:
                # concatenate distro name e.g. ubuntu-rtm
                self._distro_name = self._distro_name.replace(" ", "-")

        return self._distro_name


impl = __AptDpkgPackageInfo()



"""Apport Python module."""

import gettext
import warnings

from apport.logging import error, fatal, log, memdbg, warning
from apport.packaging_impl import impl as packaging
from apport.report import Report

__all__ = [
    "Report",
    "error",
    "fatal",
    "log",
    "memdbg",
    "packaging",
    "unicode_gettext",
    "warning",
]


def unicode_gettext(message):
    """Return the localized translation of message."""
    warnings.warn(
        "apport.unicode_gettext() is deprecated."
        " Please use gettext.gettext() directly instead.",
        PendingDeprecationWarning,
        stacklevel=2,
    )
    return gettext.gettext(message)



"""Test apport-checkreports"""

# Copyright (C) 2022 Canonical Ltd.
# Author: Benjamin Drung <bdrung@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

import os
import shutil
import subprocess
import tempfile
import unittest

import apport.report
from apport.fileutils import get_sys_uid_max
from tests.paths import get_data_directory, local_test_environment


def _is_system_user_but_not_root() -> bool:
    return os.geteuid() != 0 and os.geteuid() <= get_sys_uid_max()


class TestApportCheckreports(unittest.TestCase):
    # pylint: disable=missing-function-docstring
    """Test apport-checkreports"""

    def setUp(self):
        self.data_dir = get_data_directory()
        self.env = os.environ | local_test_environment()

        self.report_dir = tempfile.mkdtemp()
        self.env["APPORT_REPORT_DIR"] = self.report_dir

    def tearDown(self):
        shutil.rmtree(self.report_dir)

    def _call(
        self,
        args: list | None = None,
        expected_returncode: int = 0,
        expected_stdout: str = "",
    ) -> None:
        cmd = [self.data_dir / "apport-checkreports"]
        if args:
            cmd += args
        process = subprocess.run(
            cmd,
            check=False,
            env=self.env,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        self.assertEqual(process.returncode, expected_returncode)
        self.assertEqual(process.stdout, expected_stdout)
        self.assertEqual(process.stderr, "")

    def _write_report(self, filename: str, user: bool = True) -> None:
        path = f"{self.report_dir}/{filename}"
        report = apport.report.Report()
        with open(path, "wb") as report_file:
            report.write(report_file)

        if user and os.geteuid() == 0:
            os.chown(path, 1000, -1)

    @unittest.skipIf(_is_system_user_but_not_root(), "needs to run as root or user")
    def test_has_no_system_report(self):
        self._write_report("_bin_sleep.1000.crash")
        self._call(args=["--system"], expected_returncode=1)

    @unittest.skipIf(os.geteuid() != 0, "this test needs to be run as root")
    def test_has_system_report(self):
        self._write_report("_usr_bin_yes.0.crash", user=False)
        self._call(args=["-s"], expected_returncode=0, expected_stdout="yes\n")

    def test_has_user_report(self):
        self._write_report("_bin_sleep.1000.crash")
        self._call(expected_returncode=0, expected_stdout="sleep\n")

    def test_no_report(self):
        self._call(expected_returncode=1)



"""Integration tests for the apport.packaging_impl.apt_dpkg module."""

# TODO: Address following pylint complaints
# pylint: disable=invalid-name

import glob
import gzip
import os
import pathlib
import platform
import shutil
import subprocess
import tempfile
import time
import unittest

from apport.packaging_impl.apt_dpkg import WITH_DEB822_SUPPORT, impl
from tests.helper import skip_if_command_is_missing


@skip_if_command_is_missing("dpkg")
class T(unittest.TestCase):
    # pylint: disable=missing-class-docstring,missing-function-docstring
    # pylint: disable=protected-access,too-many-public-methods

    def setUp(self):
        # save and restore configuration file
        self.orig_conf = impl.configuration
        self.orig_environ = os.environ.copy()
        self.workdir = tempfile.mkdtemp()
        os.environ["HOME"] = self.workdir
        # reset internal caches between tests
        impl._apt_cache = None
        impl._sandbox_apt_cache = None

    def tearDown(self):
        impl.configuration = self.orig_conf
        os.environ.clear()
        os.environ.update(self.orig_environ)
        shutil.rmtree(self.workdir)

    def test_check_files_md5(self):
        """_check_files_md5()."""
        td = tempfile.mkdtemp()
        try:
            f1 = os.path.join(td, "test 1.txt")
            f2 = os.path.join(td, "test:2.txt")
            sumfile = os.path.join(td, "sums.txt")
            with open(f1, "w", encoding="utf-8") as fd:
                fd.write("Some stuff")
            with open(f2, "w", encoding="utf-8") as fd:
                fd.write("More stuff")
            # use one relative and one absolute path in checksums file
            with open(sumfile, "wb") as fd:
                fd.write(
                    b"2e41290da2fa3f68bd3313174467e3b5  " + f1[1:].encode() + b"\n"
                )
                fd.write(b"f6423dfbc4faf022e58b4d3f5ff71a70  " + f2.encode() + b"\n")
                fd.write(b"deadbeef000001111110000011110000  /bin/\xc3\xa4")
            self.assertEqual(impl._check_files_md5(sumfile), [], "correct md5sums")

            with open(f1, "w", encoding="utf-8") as fd:
                fd.write("Some stuff!")
            self.assertEqual(impl._check_files_md5(sumfile), [f1[1:]], "file 1 wrong")
            with open(f2, "w", encoding="utf-8") as fd:
                fd.write("More stuff!")
            self.assertEqual(
                impl._check_files_md5(sumfile), [f1[1:], f2], "files 1 and 2 wrong"
            )
            with open(f1, "w", encoding="utf-8") as fd:
                fd.write("Some stuff")
            self.assertEqual(impl._check_files_md5(sumfile), [f2], "file 2 wrong")

            # check using a direct md5 list as argument
            with open(sumfile, "rb") as fd:
                self.assertEqual(impl._check_files_md5(fd.read()), [f2], "file 2 wrong")

        finally:
            shutil.rmtree(td)

    def test_get_version(self):
        """get_version()."""
        self.assertTrue(impl.get_version("libc6").startswith("2"))
        self.assertRaises(ValueError, impl.get_version, "nonexisting")
        self.assertRaises(ValueError, impl.get_version, "wukrainian")

    def test_get_available_version(self):
        """get_available_version()."""
        self.assertTrue(impl.get_available_version("libc6").startswith("2"))
        self.assertRaises(ValueError, impl.get_available_version, "nonexisting")

    def test_get_dependencies_depends_and_pre_depends(self):
        """get_dependencies() on package with both Depends and Pre-Depends."""
        d = impl.get_dependencies("bash")
        self.assertGreater(len(d), 2)
        self.assertIn("libc6", d)
        for dep in d:
            if dep == "bash-completion":
                # bash-completion is only in Recommends (maybe not installed)
                continue
            self.assertTrue(impl.get_version(dep))

    def test_get_dependencies_pre_depends_only(self):
        """get_dependencies() on package with Pre-Depends only."""
        d = impl.get_dependencies("coreutils")
        self.assertGreaterEqual(len(d), 1)
        self.assertIn("libc6", d)
        for dep in d:
            self.assertTrue(impl.get_version(dep))

    def test_get_dependencies_depends_only(self):
        """get_dependencies() on package with Depends only."""
        d = impl.get_dependencies("sysvinit-utils")
        self.assertIn("libc6", d)
        for dep in d:
            self.assertTrue(impl.get_version(dep))

    def test_get_source(self):
        """get_source()."""
        self.assertRaises(ValueError, impl.get_source, "nonexisting")
        self.assertEqual(impl.get_source("bash"), "bash")
        self.assertIn("glibc", impl.get_source("libc6"))

    def test_get_package_origin(self):
        """get_package_origin()."""
        # determine distro name
        distro = impl.get_os_version()[0]

        self.assertRaises(ValueError, impl.get_package_origin, "nonexisting")
        # this assumes that this package is not installed
        self.assertRaises(ValueError, impl.get_package_origin, "robocode-doc")
        # this assumes that bash is native
        self.assertEqual(impl.get_package_origin("bash"), distro)
        # no non-native test here, hard to come up with a generic one

    def test_is_distro_package(self):
        """is_distro_package()."""
        self.assertRaises(ValueError, impl.is_distro_package, "nonexisting")
        self.assertTrue(impl.is_distro_package("bash"))

    def test_get_architecture(self):
        """get_architecture()."""
        self.assertRaises(ValueError, impl.get_architecture, "nonexisting")
        # just assume that bash uses the native architecture
        dpkg = subprocess.run(
            ["dpkg", "--print-architecture"],
            check=True,
            stdout=subprocess.PIPE,
            text=True,
        )
        system_arch = dpkg.stdout.strip()
        self.assertEqual(impl.get_architecture("bash"), system_arch)

    def test_get_files(self):
        """get_files()."""
        self.assertRaises(ValueError, impl.get_files, "nonexisting")
        self.assertIn("/usr/share/man/man1/bash.1.gz", impl.get_files("bash"))

    def test_get_file_package(self):
        """get_file_package() on installed files."""
        self.assertEqual(impl.get_file_package("/usr/bin/bash"), "bash")
        self.assertEqual(impl.get_file_package("/usr/bin/cat"), "coreutils")
        self.assertEqual(impl.get_file_package("/etc/pam.conf"), "libpam-runtime")
        self.assertIsNone(impl.get_file_package("/nonexisting"))

    def test_get_file_package_libc_so(self) -> None:
        """get_file_package() on libc.so.6."""
        libc_so = sorted(glob.glob("/lib/*/libc.so.6"))
        self.assertIsNotNone(libc_so)
        self.assertEqual(impl.get_file_package(libc_so[-1]), "libc6")

    def test_get_file_package_uninstalled(self):
        """get_file_package() on uninstalled packages."""
        # generate a test Contents.gz
        basedir = tempfile.mkdtemp()
        try:
            # test Contents.gz for release pocket
            mapdir = os.path.join(basedir, "dists", impl.get_distro_codename())
            os.makedirs(mapdir)
            with gzip.open(
                os.path.join(mapdir, f"Contents-{impl.get_system_architecture()}.gz"),
                "w",
            ) as f:
                f.write(
                    b"""\
usr/bin/frobnicate                                      foo/frob
usr/bin/frob                                            foo/frob-utils
usr/share/doc/frob-dev/copyright                        foo/frob-dev
bin/true                                                admin/superutils
"""
                )

            # test Contents.gz for -updates pocket
            mapdir = os.path.join(
                basedir, "dists", f"{impl.get_distro_codename()}-updates"
            )
            os.makedirs(mapdir)
            with gzip.open(
                os.path.join(mapdir, f"Contents-{impl.get_system_architecture()}.gz"),
                "w",
            ) as f:
                f.write(
                    b"lib/libnew.so.5                                        "
                    b" universe/libs/libnew5"
                )

            # use this as a mirror
            impl.set_mirror(f"file://{basedir}")

            self.assertIsNone(impl.get_file_package("usr/bin/frob", False))
            # must not match frob (same file name prefix)
            self.assertEqual(impl.get_file_package("usr/bin/frob", True), "frob-utils")
            self.assertEqual(impl.get_file_package("/usr/bin/frob", True), "frob-utils")
            # find files from -updates pocket
            self.assertIsNone(impl.get_file_package("/lib/libnew.so.5", False))
            self.assertEqual(impl.get_file_package("/lib/libnew.so.5", True), "libnew5")

            # invalid mirror
            impl.set_mirror("file:///foo/nonexisting")
            self.assertRaises(OSError, impl.get_file_package, "usr/bin/frob", True)

            # valid mirror, test cache directory
            impl.set_mirror(f"file://{basedir}")
            cache_dir = os.path.join(basedir, "cache")
            os.mkdir(cache_dir)
            self.assertEqual(
                impl.get_file_package("usr/bin/frob", True, cache_dir), "frob-utils"
            )
            cache_dir_files = sorted(os.listdir(cache_dir))
            self.assertEqual(len(cache_dir_files), 3)
            self.assertIsNone(
                impl.get_file_package(
                    "usr/share/doc/frob-dev/copyright", True, cache_dir
                )
            )

            # valid cache, should not need to access the mirror
            impl.set_mirror("file:///foo/nonexisting")
            self.assertEqual(
                impl.get_file_package("/bin/true", True, cache_dir), "superutils"
            )
            self.assertIsNone(
                impl.get_file_package(
                    "usr/share/doc/frob-dev/copyright", True, cache_dir
                )
            )
            self.assertEqual(
                impl.get_file_package("/lib/libnew.so.5", True, cache_dir), "libnew5"
            )

            # outdated cache, must refresh the cache and hit the invalid
            # mirror
            cache_file = [
                name for name in cache_dir_files if "updates" not in name
            ].pop()
            now = int(time.time())
            os.utime(os.path.join(cache_dir, cache_file), (now, now - 90000))
        finally:
            shutil.rmtree(basedir)

    def test_get_file_package_uninstalled_multiarch(self):
        """get_file_package() on foreign arches and releases"""
        # map "Foonux 3.14" to "mocky"
        orig_distro_release_to_codename = impl._distro_release_to_codename
        impl._distro_release_to_codename = (
            lambda r: (r == "Foonux 3.14") and "mocky" or None
        )

        # generate test Contents.gz for two fantasy architectures
        basedir = tempfile.mkdtemp()
        try:
            mapdir = os.path.join(basedir, "dists", impl.get_distro_codename())
            os.makedirs(mapdir)
            with gzip.open(os.path.join(mapdir, "Contents-even.gz"), "w") as f:
                f.write(
                    b"""\
usr/lib/even/libfrob.so.1                               foo/libfrob1
usr/bin/frob                                            foo/frob-utils
"""
                )
            with gzip.open(os.path.join(mapdir, "Contents-odd.gz"), "w") as f:
                f.write(
                    b"""\
usr/lib/odd/libfrob.so.1                                foo/libfrob1
usr/bin/frob                                            foo/frob-utils
"""
                )

            # and another one for fantasy release
            os.mkdir(os.path.join(basedir, "dists", "mocky"))
            with gzip.open(
                os.path.join(basedir, "dists", "mocky", "Contents-even.gz"), "w"
            ) as f:
                f.write(
                    b"""\
usr/lib/even/libfrob.so.0                               foo/libfrob0
usr/bin/frob                                            foo/frob
"""
                )

            # use this as a mirror
            impl.set_mirror(f"file://{basedir}")

            # must not match system architecture
            self.assertIsNone(impl.get_file_package("usr/bin/frob", False))
            # must match correct architecture
            self.assertEqual(
                impl.get_file_package("usr/bin/frob", True, arch="even"), "frob-utils"
            )
            self.assertEqual(
                impl.get_file_package("usr/bin/frob", True, arch="odd"), "frob-utils"
            )
            self.assertEqual(
                impl.get_file_package("/usr/lib/even/libfrob.so.1", True, arch="even"),
                "libfrob1",
            )
            self.assertEqual(
                impl.get_file_package("/usr/lib/even/libfrob.so.1", True, arch="odd"),
                None,
            )
            self.assertEqual(
                impl.get_file_package("/usr/lib/odd/libfrob.so.1", True, arch="odd"),
                "libfrob1",
            )

            # for mocky release ("Foonux 3.14")
            self.assertEqual(
                impl.get_file_package(
                    "/usr/lib/even/libfrob.so.1",
                    True,
                    release="Foonux 3.14",
                    arch="even",
                ),
                None,
            )
            self.assertEqual(
                impl.get_file_package(
                    "/usr/lib/even/libfrob.so.0",
                    True,
                    release="Foonux 3.14",
                    arch="even",
                ),
                "libfrob0",
            )
            self.assertEqual(
                impl.get_file_package(
                    "/usr/bin/frob", True, release="Foonux 3.14", arch="even"
                ),
                "frob",
            )

            # invalid mirror
            impl.set_mirror("file:///foo/nonexisting")
            self.assertRaises(
                OSError,
                impl.get_file_package,
                "/usr/lib/even/libfrob.so.1",
                True,
                arch="even",
            )
            self.assertRaises(
                OSError,
                impl.get_file_package,
                "/usr/lib/even/libfrob.so.0",
                True,
                release="Foonux 3.14",
                arch="even",
            )

            # valid mirror, test caching
            impl.set_mirror(f"file://{basedir}")
            cache_dir = os.path.join(basedir, "cache")
            os.mkdir(cache_dir)
            self.assertEqual(
                impl.get_file_package(
                    "/usr/lib/even/libfrob.so.1", True, cache_dir, arch="even"
                ),
                "libfrob1",
            )
            self.assertEqual(len(os.listdir(cache_dir)), 2)

            self.assertEqual(
                impl.get_file_package(
                    "/usr/lib/even/libfrob.so.0",
                    True,
                    cache_dir,
                    release="Foonux 3.14",
                    arch="even",
                ),
                "libfrob0",
            )
            self.assertEqual(len(os.listdir(cache_dir)), 4)

            # valid cache, should not need to access the mirror
            impl.set_mirror("file:///foo/nonexisting")
            self.assertEqual(
                impl.get_file_package("usr/bin/frob", True, cache_dir, arch="even"),
                "frob-utils",
            )
            self.assertEqual(
                impl.get_file_package(
                    "usr/bin/frob", True, cache_dir, release="Foonux 3.14", arch="even"
                ),
                "frob",
            )

            # but no cached file for the other arch
            self.assertRaises(
                OSError,
                impl.get_file_package,
                "usr/bin/frob",
                True,
                cache_dir,
                arch="odd",
            )

            # outdated cache, must refresh the cache and hit the invalid
            # mirror
            now = int(time.time())
            for cache_file in os.listdir(cache_dir):
                os.utime(os.path.join(cache_dir, cache_file), (now, now - 90000))

            self.assertRaises(
                OSError,
                impl.get_file_package,
                "usr/bin/frob",
                True,
                cache_dir,
                arch="even",
            )
        finally:
            shutil.rmtree(basedir)
            impl._distro_release_to_codename = orig_distro_release_to_codename

    def test_get_file_package_diversion(self):
        """get_file_package() for a diverted file."""
        output = subprocess.check_output(["dpkg-divert", "--list"], env={}).decode()

        for line in output.rstrip().split("\n"):
            fields = line.split(" ")
            # Local diversions have 6 fields.
            if len(fields) == 7:
                # pick first diversion we have
                break
        else:  # pragma: no cover
            self.skipTest(f"No non-local diversion found. dpkg-divert output: {output}")

        file = fields[2]
        pkg = fields[-1]

        self.assertEqual(impl.get_file_package(file), pkg)

    @unittest.skipUnless(WITH_DEB822_SUPPORT, "deb822 not supported on this system")
    def test_mirror_from_deb822_apt_sources(self) -> None:
        sources_list_d = pathlib.Path(self.workdir) / "sources.list.d"
        sources_list_d.mkdir()
        info = platform.freedesktop_os_release()
        primary_sources = sources_list_d / f"{info['ID']}.sources"
        ppa_sources = sources_list_d / "0ppa.sources"
        primary_sources.write_text(
            """\
Types: deb
URIs: http://primary-mirror.example.com/distro/
Suites: tuxy
Components: main
"""
        )
        ppa_sources.write_text(
            """\
Types: deb
URIs: https://ppa.example.net/user/ppa-name/distro/
Suites: lazy
Components: main
"""
        )

        self.assertEqual(
            impl._get_primary_mirror_from_apt_sources(self.workdir),
            "http://primary-mirror.example.com/distro/",
        )

    def test_mirror_from_apt_sources(self):
        s = os.path.join(self.workdir, "sources.list")

        # valid file, should grab the first mirror
        with open(s, "w", encoding="utf-8") as f:
            f.write(
                """# some comment
deb-src http://source.mirror/foo tuxy main
deb http://binary.mirror/tuxy tuxy main
deb http://secondary.mirror tuxy extra
"""
            )
            f.flush()
            self.assertEqual(
                impl._get_primary_mirror_from_apt_sources(self.workdir),
                "http://binary.mirror/tuxy",
            )

        # valid file with options
        with open(s, "w", encoding="utf-8") as f:
            f.write(
                """# some comment
deb-src http://source.mirror/foo tuxy main
deb [arch=flowerpc,leghf] http://binary.mirror/tuxy tuxy main
deb http://secondary.mirror tuxy extra
"""
            )
            f.flush()
            self.assertEqual(
                impl._get_primary_mirror_from_apt_sources(self.workdir),
                "http://binary.mirror/tuxy",
            )

        # empty file
        with open(s, "w", encoding="utf-8") as f:
            f.flush()
        self.assertRaises(
            SystemError, impl._get_primary_mirror_from_apt_sources, self.workdir
        )

    def test_mirror_from_apt_sources_with_options(self) -> None:
        """Test _get_primary_mirror_from_apt_sources() with options"""
        sources = pathlib.Path(self.workdir) / "sources.list"
        expected = "https://example.com/"
        sources.write_text(f"deb [ arch=riscv64 ] {expected} suite component")
        actual = impl._get_primary_mirror_from_apt_sources(self.workdir)
        self.assertEqual(actual, expected)

    def test_get_modified_conffiles(self):
        """get_modified_conffiles()"""
        # very shallow
        self.assertEqual(type(impl.get_modified_conffiles("bash")), type({}))
        self.assertEqual(type(impl.get_modified_conffiles("apport")), type({}))
        self.assertEqual(type(impl.get_modified_conffiles("nonexisting")), type({}))

    def test_get_system_architecture(self):
        """get_system_architecture()."""
        arch = impl.get_system_architecture()
        # must be nonempty without line breaks
        self.assertNotEqual(arch, "")
        self.assertNotIn("\n", arch)

    def test_get_library_paths(self):
        """get_library_paths()."""
        paths = impl.get_library_paths()
        # must be nonempty without line breaks
        self.assertNotEqual(paths, "")
        self.assertIn(":", paths)
        self.assertIn("/lib", paths)
        self.assertNotIn("\n", paths)

    def test_compare_versions(self):
        """compare_versions."""
        self.assertEqual(impl.compare_versions("1", "2"), -1)
        self.assertEqual(impl.compare_versions("1.0-1ubuntu1", "1.0-1ubuntu2"), -1)
        self.assertEqual(impl.compare_versions("1.0-1ubuntu1", "1.0-1ubuntu1"), 0)
        self.assertEqual(impl.compare_versions("1.0-1ubuntu2", "1.0-1ubuntu1"), 1)
        self.assertEqual(impl.compare_versions("1:1.0-1", "2007-2"), 1)
        self.assertEqual(impl.compare_versions("1:1.0-1~1", "1:1.0-1"), -1)

    def test_enabled(self):
        """enabled."""
        impl.configuration = "/nonexisting"
        self.assertEqual(impl.enabled(), True)

        with tempfile.NamedTemporaryFile() as f:
            impl.configuration = f.name
            f.write("# configuration file\nenabled = 1".encode())
            f.flush()
            self.assertEqual(impl.enabled(), True)

        with tempfile.NamedTemporaryFile() as f:
            impl.configuration = f.name
            f.write("# configuration file\n  enabled =0  ".encode())
            f.flush()
            self.assertEqual(impl.enabled(), False)

        with tempfile.NamedTemporaryFile() as f:
            impl.configuration = f.name
            f.write("# configuration file\nnothing here".encode())
            f.flush()
            self.assertEqual(impl.enabled(), True)

    def test_get_kernel_package(self):
        """get_kernel_package()."""
        self.assertIn("linux", impl.get_kernel_package())

    def test_package_name_glob(self):
        """package_name_glob()."""
        self.assertGreater(len(impl.package_name_glob("a*")), 5)
        self.assertIn("bash", impl.package_name_glob("ba*h"))
        self.assertEqual(impl.package_name_glob("bash"), ["bash"])
        self.assertEqual(impl.package_name_glob("xzywef*"), [])



"""Integration tests for the apport.packaging_impl.rpm module."""

import unittest

from tests.helper import skip_if_command_is_missing

try:
    from apport.packaging_impl.rpm import impl

    HAS_RPM = True
except ImportError:
    HAS_RPM = False


@unittest.skipUnless(HAS_RPM, "rpm module not available")
@skip_if_command_is_missing("rpm")
class TestPackagingRpm(unittest.TestCase):
    """Integration tests for the apport.packaging_impl.rpm module."""

    def test_get_dependencies(self):
        """get_dependencies()."""
        deps = impl.get_dependencies("bash")
        self.assertNotEqual(deps, [])

    def test_get_header(self):
        """_get_header()."""
        # pylint: disable=protected-access
        hdr = impl._get_header("alsa-utils")
        self.assertEqual(hdr["n"], "alsa-utils")

    def test_get_headers_by_tag(self):
        """_get_headers_by_tag()."""
        # pylint: disable=protected-access
        headers_by_tag = impl._get_headers_by_tag("basenames", "/bin/bash")
        self.assertEqual(len(headers_by_tag), 1)
        self.assertTrue(headers_by_tag[0]["n"].startswith("bash"))

    def test_get_system_architecture(self):
        """get_system_architecture()."""
        arch = impl.get_system_architecture()
        # must be nonempty without line breaks
        self.assertNotEqual(arch, "")
        self.assertNotIn("\n", arch)

    def test_get_version(self):
        """get_version()."""
        ver = impl.get_version("bash")
        self.assertIsNotNone(ver)
        ver = impl.get_version("alsa-utils")
        self.assertIsNotNone(ver)



"""Detect if the current session is running under Wayland."""

import os


def add_info(report, unused_ui):
    """Add a tag if current session is running under Wayland."""
    if os.environ.get("WAYLAND_DISPLAY"):
        report.add_tags(["wayland-session"])



"""Integration tests for the GTK Apport user interface."""

# Copyright (C) 2025 Canonical Ltd.
# Author: Benjamin Drung <benjamin.drung@canonical.com>
# SPDX-License-Identifier: GPL-2.0-or-later

import io
import os
import unittest.mock

import pytest

from tests.helper import import_module_from_file, restore_os_environ
from tests.paths import get_data_directory


@restore_os_environ()
def test_unusable_display() -> None:
    """Test apport-gtk to not crash if no usable display is found (LP: #2006981)."""
    os.environ |= {"DISPLAY": ":42", "WAYLAND_DISPLAY": "bogus"}
    apport_gtk_path = get_data_directory("gtk") / "apport-gtk"
    with unittest.mock.patch("sys.stderr", new_callable=io.StringIO) as stderr:
        try:
            apport_gtk = import_module_from_file(apport_gtk_path)
        except SystemExit:
            pytest.skip(stderr.getvalue().strip())
    with unittest.mock.patch("sys.stderr", new_callable=io.StringIO) as stderr:
        with pytest.raises(SystemExit) as error:
            apport_gtk.main([str(apport_gtk_path)])

    assert (
        stderr.getvalue() == "ERROR: This program needs a running display server"
        ' session. Please see "man apport-cli" for a command line version of Apport.\n'
    )
    assert error.value.code == 1



"""GTK Apport user interface tests."""

# Copyright (C) 2012 Canonical Ltd.
# Author: Evan Dandrea <evan.dandrea@canonical.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

# pylint: disable=too-many-lines
# TODO: Address following pylint complaints
# pylint: disable=invalid-name

import os
import shutil
import subprocess
import tempfile
import textwrap
import unittest
import unittest.mock
from gettext import gettext as _
from unittest.mock import MagicMock

import apport.crashdb_impl.memory
import apport.report
from tests.helper import import_module_from_file
from tests.paths import get_data_directory, local_test_environment

try:
    import gi

    gi.require_version("Gtk", "3.0")  # noqa: E402, pylint: disable=C0413
    from gi.repository import GLib, GObject, Gtk

    GLib.log_set_always_fatal(
        GLib.LogLevelFlags.LEVEL_WARNING | GLib.LogLevelFlags.LEVEL_CRITICAL
    )

    GI_IMPORT_ERROR = None
except ImportError as error:
    GI_IMPORT_ERROR = error

apport_gtk_path = get_data_directory("gtk") / "apport-gtk"
if not GI_IMPORT_ERROR:
    apport_gtk = import_module_from_file(apport_gtk_path)
    GTKUserInterface = apport_gtk.GTKUserInterface
else:
    apport_gtk = None
    GTKUserInterface = None


@unittest.skipIf(GI_IMPORT_ERROR, f"gi Python module not available: {GI_IMPORT_ERROR}")
class T(unittest.TestCase):  # pylint: disable=too-many-public-methods
    # pylint: disable=missing-class-docstring,missing-function-docstring
    POLLING_INTERVAL_MS = 10

    @classmethod
    def setUpClass(cls):
        cls.orig_environ = os.environ.copy()
        os.environ |= local_test_environment()
        os.environ["LANGUAGE"] = "C"
        r = apport.report.Report()
        r.add_os_info()
        cls.distro = r["DistroRelease"].split()[0]

    @classmethod
    def tearDownClass(cls):
        os.environ.clear()
        os.environ.update(cls.orig_environ)

    def setUp(self):
        self.report_dir = tempfile.mkdtemp()
        apport.fileutils.report_dir = self.report_dir
        os.environ["APPORT_REPORT_DIR"] = self.report_dir
        # do not cause eternal hangs because of error dialog boxes
        os.environ["APPORT_IGNORE_OBSOLETE_PACKAGES"] = "1"
        os.environ["APPORT_DISABLE_DISTRO_CHECK"] = "1"

        self.app = GTKUserInterface([str(apport_gtk_path)])

        # use in-memory crashdb
        self.app.crashdb = apport.crashdb_impl.memory.CrashDatabase(None, {})

        # test report
        self.app.report_file = os.path.join(self.report_dir, "bash.crash")

        self.app.report = apport.report.Report()
        self.app.report["ExecutablePath"] = "/bin/bash"
        self.app.report["Signal"] = "11"
        self.app.report["CoreDump"] = b"\x01\x02"
        self.app.report["DistroRelease"] = self.distro
        with open(self.app.report_file, "wb") as f:
            self.app.report.write(f)

        # disable package hooks, as they might ask for sudo password and other
        # interactive bits; allow tests to install their own hooks
        self.hook_dir = tempfile.mkdtemp()
        apport.report.GENERAL_HOOK_DIR = self.hook_dir
        apport.report.PACKAGE_HOOK_DIR = self.hook_dir

    def tearDown(self):
        shutil.rmtree(self.report_dir)
        shutil.rmtree(self.hook_dir)

    def test_close_button(self):
        """Clicking the close button on the window does not report the
        crash."""

        def c():
            self.app.w("dialog_crash_new").destroy()

        GLib.idle_add(c)
        result = self.app.ui_present_report_details(True)
        self.assertFalse(result.report)

    def test_kernel_crash_layout(self):
        """Display crash dialog for kernel crash.

        +-----------------------------------------------------------------+
        | [ logo] YourDistro has experienced an internal error.           |
        |            Send problem report to the developers?               |
        |                                                                 |
        |            [ ] Remember this in future                          |
        |                                                                 |
        | [ Show Details ]                      [ Don't send ]   [ Send ] |
        +-----------------------------------------------------------------+
        """
        self.app.report["ProblemType"] = "KernelCrash"
        GLib.idle_add(Gtk.main_quit)
        self.app.ui_present_report_details(True)
        self.assertEqual(self.app.w("dialog_crash_new").get_title(), self.distro)
        self.assertEqual(
            self.app.w("title_label").get_text(),
            _("Sorry, %s has experienced an internal error.") % self.distro,
        )
        self.assertTrue(self.app.w("send_problem_notice_label").get_property("visible"))
        remember_send_error_report = self.app.w("remember_send_report_choice")
        self.assertTrue(remember_send_error_report.get_property("visible"))
        self.assertFalse(remember_send_error_report.get_active())
        self.assertTrue(self.app.w("show_details").get_property("visible"))
        self.assertTrue(self.app.w("dont_send_button").get_property("visible"))
        self.assertTrue(self.app.w("continue_button").get_property("visible"))
        self.assertEqual(self.app.w("continue_button").get_label(), _("Send"))
        self.assertFalse(self.app.w("subtitle_label").get_property("visible"))
        self.assertFalse(self.app.w("ignore_future_problems").get_property("visible"))

    def test_package_crash_layout(self):
        """Display crash dialog for a failed package installation.

        +-----------------------------------------------------------------+
        | [ error  ] Sorry, a problem occurred while installing software. |
        |            Send problem report to the developers?               |
        |            Package: apport 1.2.3~0ubuntu1                       |
        |                                                                 |
        |            [ ] Remember this in future                          |
        |                                                                 |
        | [ Show Details ]                      [ Don't send ]   [ Send ] |
        +-----------------------------------------------------------------+
        """
        self.app.report["ProblemType"] = "Package"
        self.app.report["Package"] = "apport 1.2.3~0ubuntu1"
        GLib.idle_add(Gtk.main_quit)
        self.app.ui_present_report_details(True)
        self.assertEqual(self.app.w("dialog_crash_new").get_title(), self.distro)
        self.assertEqual(
            self.app.w("title_label").get_text(),
            _("Sorry, a problem occurred while installing software."),
        )
        self.assertTrue(self.app.w("send_problem_notice_label").get_property("visible"))
        remember_send_error_report = self.app.w("remember_send_report_choice")
        self.assertTrue(remember_send_error_report.get_property("visible"))
        self.assertFalse(remember_send_error_report.get_active())
        self.assertTrue(self.app.w("show_details").get_property("visible"))
        self.assertTrue(self.app.w("dont_send_button").get_property("visible"))
        self.assertTrue(self.app.w("continue_button").get_property("visible"))
        self.assertEqual(self.app.w("continue_button").get_label(), _("Send"))
        self.assertTrue(self.app.w("subtitle_label").get_property("visible"))
        self.assertEqual(
            self.app.w("subtitle_label").get_text(), _("Package: apport 1.2.3~0ubuntu1")
        )

    def test_regular_crash_thread_layout(self):
        """A thread of execution has failed, but the application persists.

        +-----------------------------------------------------------------+
        | [ logo] YourDistro has experienced an internal error.           |
        |            Send problem report to the developers?               |
        |            If you notice further problems, try restarting the   |
        |            computer.                                            |
        |                                                                 |
        |            [ ] Remember this in future                          |
        |                                                                 |
        | [ Show Details ]                      [ Don't send ]   [ Send ] |
        +-----------------------------------------------------------------+
        """
        self.app.report["ProblemType"] = "Crash"
        self.app.report["ProcStatus"] = "Name:\tsystemd\nPid:\t1"
        GLib.idle_add(Gtk.main_quit)
        self.app.ui_present_report_details(True)
        self.assertTrue(self.app.w("dont_send_button").get_property("visible"))
        self.assertEqual(self.app.w("continue_button").get_label(), _("Send"))

    def test_regular_crash_layout(self):
        """Display crash dialog for an application crash.

        +-----------------------------------------------------------------+
        | [ apport ] The application Apport has closed unexpectedly.      |
        |            Send problem report to the developers?               |
        |                                                                 |
        |            [ ] Remember this in future                          |
        |            [ ] Ignore future problems of this program version   |
        |                                                                 |
        | [ Show Details ]                      [ Don't send ]   [ Send ] |
        +-----------------------------------------------------------------+
        """
        self.app.report["ProblemType"] = "Crash"
        self.app.report["CrashCounter"] = "1"
        self.app.report["Package"] = "apport 1.2.3~0ubuntu1"
        with tempfile.NamedTemporaryFile(mode="w+") as fp:
            fp.write(
                textwrap.dedent(
                    """\
                    [Desktop Entry]
                    Version=1.0
                    Name=Apport
                    Type=Application
                    """
                )
            )
            fp.flush()
            self.app.report["DesktopFile"] = fp.name
            GLib.idle_add(Gtk.main_quit)
            self.app.ui_present_report_details(True)
        self.assertEqual(self.app.w("dialog_crash_new").get_title(), self.distro)
        self.assertEqual(
            self.app.w("title_label").get_text(),
            _("The application Apport has closed unexpectedly."),
        )
        self.assertTrue(self.app.w("send_problem_notice_label").get_property("visible"))
        remember_send_error_report = self.app.w("remember_send_report_choice")
        self.assertTrue(remember_send_error_report.get_property("visible"))
        self.assertFalse(remember_send_error_report.get_active())
        self.assertTrue(self.app.w("show_details").get_property("visible"))
        self.assertTrue(self.app.w("dont_send_button").get_property("visible"))
        # no ProcCmdline, cannot restart
        self.assertTrue(self.app.w("continue_button").get_property("visible"))
        self.assertEqual(self.app.w("continue_button").get_label(), _("Send"))
        self.assertFalse(self.app.w("subtitle_label").get_property("visible"))
        self.assertTrue(self.app.w("ignore_future_problems").get_property("visible"))
        self.assertTrue(
            self.app.w("ignore_future_problems")
            .get_label()
            .endswith("of this program version")
        )

    def test_regular_crash_layout_restart(self):
        """Display crash dialog for an application crash offering a restart.

        +-----------------------------------------------------------------+
        | [ apport ] The application Apport has closed unexpectedly.      |
        |            Send problem report to the developers?               |
        |                                                                 |
        |            [ ] Remember this in future                          |
        |            [ ] Ignore future problems of this program version   |
        |            [X] Relaunch this application                        |
        |                                                                 |
        | [ Show Details ]                      [ Don't send ]   [ Send ] |
        +-----------------------------------------------------------------+
        """
        # pretend we got called through run_crashes() which sets offer_restart
        self.app.offer_restart = True
        self.app.report["ProblemType"] = "Crash"
        self.app.report["CrashCounter"] = "1"
        self.app.report["ProcCmdline"] = "apport-bug apport"
        self.app.report["Package"] = "apport 1.2.3~0ubuntu1"
        with tempfile.NamedTemporaryFile(mode="w+") as fp:
            fp.write(
                textwrap.dedent(
                    """\
                    [Desktop Entry]
                    Version=1.0
                    Name=Apport
                    Type=Application
                    """
                )
            )
            fp.flush()
            self.app.report["DesktopFile"] = fp.name
            GLib.idle_add(Gtk.main_quit)
            self.app.ui_present_report_details(True)
        self.assertEqual(self.app.w("dialog_crash_new").get_title(), self.distro)
        self.assertEqual(
            self.app.w("title_label").get_text(),
            _("The application Apport has closed unexpectedly."),
        )
        self.assertTrue(self.app.w("send_problem_notice_label").get_property("visible"))
        remember_send_error_report = self.app.w("remember_send_report_choice")
        self.assertTrue(remember_send_error_report.get_property("visible"))
        self.assertFalse(remember_send_error_report.get_active())
        self.assertTrue(self.app.w("show_details").get_property("visible"))
        self.assertTrue(self.app.w("dont_send_button").get_property("visible"))
        self.assertTrue(self.app.w("continue_button").get_property("visible"))
        self.assertEqual(self.app.w("continue_button").get_label(), _("Send"))
        self.assertFalse(self.app.w("subtitle_label").get_property("visible"))
        self.assertTrue(self.app.w("ignore_future_problems").get_property("visible"))
        self.assertTrue(
            self.app.w("ignore_future_problems")
            .get_label()
            .endswith("of this program version")
        )
        self.assertTrue(self.app.w("relaunch_app").get_property("visible"))
        self.assertTrue(self.app.w("relaunch_app").get_active())

    def test_regular_crash_layout_norestart(self):
        """Display crash dialog for an application crash offering no restart.

        +-----------------------------------------------------------------+
        | [ apport ] The application Apport has closed unexpectedly.      |
        |            Send problem report to the developers?               |
        |                                                                 |
        |            [ ] Remember this in future                          |
        |            [ ] Ignore future problems of this program version.  |
        |                                                                 |
        | [ Show Details ]                      [ Don't send ]   [ Send ] |
        +-----------------------------------------------------------------+
        """
        # pretend we did not get called through run_crashes(), thus
        # no offer_restart
        self.app.report["ProblemType"] = "Crash"
        self.app.report["CrashCounter"] = "1"
        self.app.report["ProcCmdline"] = "apport-bug apport"
        self.app.report["Package"] = "apport 1.2.3~0ubuntu1"
        with tempfile.NamedTemporaryFile(mode="w+") as fp:
            fp.write(
                textwrap.dedent(
                    """\
                    [Desktop Entry]
                    Version=1.0
                    Name=Apport
                    Type=Application
                    """
                )
            )
            fp.flush()
            self.app.report["DesktopFile"] = fp.name
            GLib.idle_add(Gtk.main_quit)
            self.app.ui_present_report_details(True)
        self.assertEqual(self.app.w("dialog_crash_new").get_title(), self.distro)
        self.assertEqual(
            self.app.w("title_label").get_text(),
            _("The application Apport has closed unexpectedly."),
        )
        self.assertTrue(self.app.w("send_problem_notice_label").get_property("visible"))
        remember_send_error_report = self.app.w("remember_send_report_choice")
        self.assertTrue(remember_send_error_report.get_property("visible"))
        self.assertFalse(remember_send_error_report.get_active())
        self.assertTrue(self.app.w("show_details").get_property("visible"))
        self.assertTrue(self.app.w("dont_send_button").get_property("visible"))
        self.assertTrue(self.app.w("continue_button").get_property("visible"))
        self.assertEqual(self.app.w("continue_button").get_label(), _("Send"))

    def test_hang_layout(self):
        """Display crash dialog for a hanging process.

        +-----------------------------------------------------------------+
        | [ apport ] The application Apport has stopped responding.       |
        |            Send problem report to the developers?               |
        |            You can wait to see if it wakes up, or close, or     |
        |            relaunch it.                                         |
        |                                                                 |
        |            [ ] Remember this in future                          |
        |            [X] Relaunch this application                        |
        |                                                                 |
        | [ Show Details ]                      [ Don't send ]   [ Send ] |
        +-----------------------------------------------------------------+
        """
        # pretend we got called through run_crashes() which sets offer_restart
        self.app.offer_restart = True
        self.app.report["ProblemType"] = "Hang"
        self.app.report["ProcCmdline"] = "apport-bug apport"
        self.app.report["Package"] = "apport 1.2.3~0ubuntu1"
        with tempfile.NamedTemporaryFile(mode="w+") as fp:
            fp.write(
                textwrap.dedent(
                    """\
                    [Desktop Entry]
                    Version=1.0
                    Name=Apport
                    Type=Application
                    """
                )
            )
            fp.flush()
            self.app.report["DesktopFile"] = fp.name
            GLib.idle_add(Gtk.main_quit)
            self.app.ui_present_report_details(True)
        self.assertEqual(self.app.w("dialog_crash_new").get_title(), self.distro)
        self.assertEqual(
            self.app.w("title_label").get_text(),
            _("The application Apport has stopped responding."),
        )
        self.assertEqual(
            self.app.w("subtitle_label").get_text(),
            _("You can wait to see if it wakes up, or close or relaunch it."),
        )
        self.assertTrue(self.app.w("send_problem_notice_label").get_property("visible"))
        remember_send_error_report = self.app.w("remember_send_report_choice")
        self.assertTrue(remember_send_error_report.get_property("visible"))
        self.assertFalse(remember_send_error_report.get_active())
        self.assertTrue(self.app.w("show_details").get_property("visible"))
        self.assertTrue(self.app.w("dont_send_button").get_property("visible"))
        self.assertTrue(self.app.w("continue_button").get_property("visible"))
        self.assertEqual(self.app.w("continue_button").get_label(), _("Send"))
        self.assertTrue(self.app.w("subtitle_label").get_property("visible"))
        self.assertFalse(self.app.w("ignore_future_problems").get_property("visible"))

    def test_system_crash_layout(self):
        """Display crash dialog for a system application crash.

        +---------------------------------------------------------------+
        | [ logo ] Sorry, YourDistro has experienced an internal error. |
        |          Send problem report to the developers?               |
        |          If you notice further problems, try restarting the   |
        |          computer                                             |
        |                                                               |
        |            [ ] Remember this in future                        |
        |            [ ] Ignore future problems of this type.           |
        |                                                               |
        | [ Show Details ]                    [ Don't send ]   [ Send ] |
        +---------------------------------------------------------------+
        """
        self.app.report["ProblemType"] = "Crash"
        self.app.report["CrashCounter"] = "1"
        self.app.report["Package"] = "bash 5"
        GLib.idle_add(Gtk.main_quit)
        self.app.ui_present_report_details(True)
        self.assertEqual(self.app.w("dialog_crash_new").get_title(), self.distro)
        self.assertEqual(
            self.app.w("title_label").get_text(),
            _("Sorry, %s has experienced an internal error.") % self.distro,
        )
        self.assertEqual(
            self.app.w("subtitle_label").get_text(),
            _("If you notice further problems, try restarting the computer."),
        )
        self.assertTrue(self.app.w("subtitle_label").get_property("visible"))
        self.assertTrue(self.app.w("send_problem_notice_label").get_property("visible"))
        remember_send_error_report = self.app.w("remember_send_report_choice")
        self.assertTrue(remember_send_error_report.get_property("visible"))
        self.assertFalse(remember_send_error_report.get_active())
        self.assertTrue(self.app.w("show_details").get_property("visible"))
        self.assertTrue(self.app.w("dont_send_button").get_property("visible"))
        self.assertTrue(self.app.w("continue_button").get_property("visible"))
        self.assertEqual(self.app.w("continue_button").get_label(), _("Send"))
        self.assertTrue(self.app.w("ignore_future_problems").get_property("visible"))
        self.assertTrue(
            self.app.w("ignore_future_problems").get_label().endswith("of this type")
        )

    def test_system_crash_from_console_layout(self):
        """Display crash dialog from console.

        +-------------------------------------------------------------------+
        | [ ubuntu ] Sorry, the application apport has closed unexpectedly. |
        |            Send problem report to the developers?                 |
        |            If you notice further problems, try restarting the     |
        |            computer                                               |
        |                                                                   |
        |            [ ] Remember this in future                            |
        |                                                                   |
        | [ Show Details ]                      [ Don't send ]   [ Send ]   |
        +-------------------------------------------------------------------+
        """
        self.app.report["ProblemType"] = "Crash"
        self.app.report["Package"] = "bash 5"
        self.app.report["ProcEnviron"] = "LANGUAGE=en_GB:en\nSHELL=/bin/sh\nTERM=xterm"
        self.app.report["ExecutablePath"] = "/usr/bin/apport"
        # This will be set by apport/ui.py in load_report()
        self.app.cur_package = "apport"
        GLib.idle_add(Gtk.main_quit)
        self.app.ui_present_report_details(True)
        self.assertEqual(self.app.w("dialog_crash_new").get_title(), self.distro)
        self.assertEqual(
            self.app.w("title_label").get_text(),
            _("Sorry, the application apport has stopped unexpectedly."),
        )
        self.assertEqual(
            self.app.w("subtitle_label").get_text(),
            _("If you notice further problems, try restarting the computer."),
        )
        self.assertTrue(self.app.w("subtitle_label").get_property("visible"))
        self.assertTrue(self.app.w("send_problem_notice_label").get_property("visible"))
        remember_send_error_report = self.app.w("remember_send_report_choice")
        self.assertTrue(remember_send_error_report.get_property("visible"))
        self.assertFalse(remember_send_error_report.get_active())
        self.assertTrue(self.app.w("show_details").get_property("visible"))
        self.assertTrue(self.app.w("dont_send_button").get_property("visible"))
        self.assertTrue(self.app.w("continue_button").get_property("visible"))
        self.assertEqual(self.app.w("continue_button").get_label(), _("Send"))

        del self.app.report["ExecutablePath"]
        GLib.idle_add(Gtk.main_quit)
        self.app.ui_present_report_details(True)
        self.assertEqual(
            self.app.w("title_label").get_text(),
            _("Sorry, apport has closed unexpectedly."),
        )

        # no crash counter
        self.assertFalse(self.app.w("ignore_future_problems").get_property("visible"))

    @unittest.mock.patch.object(GTKUserInterface, "can_examine_locally", MagicMock())
    def test_examine_button(self):
        """Crash dialog showing or not showing "Examine locally".

        +---------------------------------------------------------------------+
        | [ apport ] The application Apport has closed unexpectedly.          |
        |            Send problem report to the developers?                   |
        |                                                                     |
        |            [ ] Remember this in future                              |
        |                                                                     |
        | [ Show Details ] [ Examine locally ]      [ Don't send ]   [ Send ] |
        +---------------------------------------------------------------------+
        """
        self.app.report["ProblemType"] = "Crash"
        self.app.report["Package"] = "bash 5"

        GLib.idle_add(Gtk.main_quit)
        self.app.can_examine_locally.return_value = False
        self.app.ui_present_report_details(True)
        self.assertFalse(self.app.w("examine").get_property("visible"))

        GLib.idle_add(self.app.w("examine").clicked)
        self.app.can_examine_locally.return_value = True
        result = self.app.ui_present_report_details(True)
        self.assertTrue(self.app.w("examine").get_property("visible"))
        self.assertTrue(result.examine)

    def test_apport_bug_package_layout(self):
        """Display report detail dialog.

        +-------------------------------------------------------------------+
        | [ error  ] Send problem report to the developers?                 |
        |                                                                   |
        |            +----------------------------------------------------+ |
        |            | |> ApportVersion                                   | |
        |            | ...                                                | |
        |            +----------------------------------------------------+ |
        |                                                                   |
        |                                         [ Don't send ]   [ Send ] |
        +-------------------------------------------------------------------+
        """
        self.app.report_file = None
        GLib.idle_add(Gtk.main_quit)
        self.app.ui_present_report_details(True)
        self.assertEqual(
            self.app.w("title_label").get_text(),
            _("Send problem report to the developers?"),
        )
        self.assertFalse(self.app.w("subtitle_label").get_property("visible"))
        self.assertFalse(
            self.app.w("send_problem_notice_label").get_property("visible")
        )
        remember_send_error_report = self.app.w("remember_send_report_choice")
        self.assertFalse(remember_send_error_report.get_property("visible"))
        self.assertFalse(remember_send_error_report.get_active())
        self.assertFalse(self.app.w("show_details").get_property("visible"))
        self.assertTrue(self.app.w("continue_button").get_property("visible"))
        self.assertEqual(self.app.w("continue_button").get_label(), _("Send"))
        self.assertTrue(self.app.w("dont_send_button").get_property("visible"))
        self.assertTrue(self.app.w("details_scrolledwindow").get_property("visible"))
        self.assertTrue(self.app.w("dialog_crash_new").get_resizable())

    def test_apport_bug_package_layout_load_file(self):
        """Bug layout from a loaded report."""
        self.app.report_file = "/tmp/foo.apport"
        self.app.report = apport.report.Report("Bug")
        self.app.report["Package"] = "libfoo1"
        self.app.report["SourcePackage"] = "foo"

        GLib.idle_add(Gtk.main_quit)
        self.app.ui_present_report_details(True)
        self.assertEqual(
            self.app.w("title_label").get_text(),
            _("Send problem report to the developers?"),
        )
        self.assertFalse(self.app.w("subtitle_label").get_property("visible"))
        self.assertFalse(
            self.app.w("send_problem_notice_label").get_property("visible")
        )
        remember_send_error_report = self.app.w("remember_send_report_choice")
        self.assertFalse(remember_send_error_report.get_property("visible"))
        self.assertFalse(remember_send_error_report.get_active())
        self.assertFalse(self.app.w("show_details").get_property("visible"))
        self.assertTrue(self.app.w("continue_button").get_property("visible"))
        self.assertEqual(self.app.w("continue_button").get_label(), _("Send"))
        self.assertTrue(self.app.w("dont_send_button").get_property("visible"))
        self.assertTrue(self.app.w("details_scrolledwindow").get_property("visible"))
        self.assertTrue(self.app.w("dialog_crash_new").get_resizable())

    def test_recoverable_crash_layout(self):
        """Display crash dialog for a recoverable crash.

        +-----------------------------------------------------------------+
        | [ logo ] The application Foo has experienced an internal error. |
        |            Send problem report to the developers?               |
        |          Developer-specified error text.                        |
        |                                                                 |
        |            [ ] Remember this in future                          |
        |                                                                 |
        | [ Show Details ]                      [ Don't send ]   [ Send ] |
        +-----------------------------------------------------------------+
        """
        self.app.report["ProblemType"] = "RecoverableProblem"
        self.app.report["Package"] = "apport 1.2.3~0ubuntu1"
        self.app.report["DialogBody"] = "Some developer-specified error text."
        with tempfile.NamedTemporaryFile(mode="w+") as fp:
            fp.write(
                textwrap.dedent(
                    """\
                    [Desktop Entry]
                    Version=1.0
                    Name=Apport
                    Type=Application
                    """
                )
            )
            fp.flush()
            self.app.report["DesktopFile"] = fp.name
            GLib.idle_add(Gtk.main_quit)
            self.app.ui_present_report_details(True)
        self.assertEqual(self.app.w("dialog_crash_new").get_title(), self.distro)
        msg = "The application Apport has experienced an internal error."
        self.assertEqual(self.app.w("title_label").get_text(), msg)
        msg = "Some developer-specified error text."
        self.assertEqual(self.app.w("subtitle_label").get_text(), msg)
        self.assertTrue(self.app.w("subtitle_label").get_property("visible"))
        self.assertTrue(self.app.w("send_problem_notice_label").get_property("visible"))
        remember_send_error_report = self.app.w("remember_send_report_choice")
        self.assertTrue(remember_send_error_report.get_property("visible"))
        self.assertFalse(remember_send_error_report.get_active())
        self.assertTrue(self.app.w("show_details").get_property("visible"))
        self.assertTrue(self.app.w("continue_button").get_property("visible"))
        self.assertEqual(self.app.w("continue_button").get_label(), _("Send"))

    def test_administrator_disabled_reporting(self):
        GLib.idle_add(Gtk.main_quit)
        self.app.ui_present_report_details(False)
        self.assertFalse(
            self.app.w("send_problem_notice_label").get_property("visible")
        )
        remember_send_error_report = self.app.w("remember_send_report_choice")
        self.assertFalse(remember_send_error_report.get_property("visible"))
        self.assertFalse(remember_send_error_report.get_active())

    @unittest.mock.patch.object(GTKUserInterface, "open_url", MagicMock())
    @unittest.mock.patch.object(
        GTKUserInterface, "ui_start_upload_progress", MagicMock()
    )
    @unittest.mock.patch.object(
        GTKUserInterface, "ui_stop_upload_progress", MagicMock()
    )
    @unittest.mock.patch("apport.report.Report.add_gdb_info", MagicMock())
    @unittest.mock.patch(
        "apport.fileutils.allowed_to_report", MagicMock(return_value=True)
    )
    def test_crash_nodetails(self):
        """Crash report without showing details"""

        def cont():
            if Gtk.events_pending():
                return True  # pragma: no cover
            if not self.app.w("continue_button").get_visible():
                return True  # pragma: no cover
            self.app.w("continue_button").clicked()
            return False

        GLib.timeout_add(self.POLLING_INTERVAL_MS, cont)
        info_collection_window = self.app.w("window_information_collection")
        with unittest.mock.patch.object(
            info_collection_window, "show", wraps=info_collection_window.show
        ) as info_collection_window_show_mock:
            self.app.run_crash(self.app.report_file)

        # we should have reported one crash
        self.assertEqual(self.app.crashdb.latest_id(), 0)
        r = self.app.crashdb.download(0)
        self.assertEqual(r["ProblemType"], "Crash")
        self.assertEqual(r["ExecutablePath"], "/bin/bash")

        # should show a progress bar for info collection
        info_collection_window_show_mock.assert_called_once_with()

        # data was collected
        self.assertTrue(r["Package"].startswith("bash "))
        self.assertIn("libc", r["Dependencies"])

        # upload dialog shown
        self.assertEqual(self.app.ui_start_upload_progress.call_count, 1)
        self.assertEqual(self.app.ui_stop_upload_progress.call_count, 1)

        # URL was opened
        self.assertEqual(self.app.open_url.call_count, 1)

    @unittest.mock.patch.object(GTKUserInterface, "open_url", MagicMock())
    @unittest.mock.patch.object(
        GTKUserInterface, "ui_start_upload_progress", MagicMock()
    )
    @unittest.mock.patch.object(
        GTKUserInterface, "ui_stop_upload_progress", MagicMock()
    )
    @unittest.mock.patch("apport.report.Report.add_gdb_info", MagicMock())
    @unittest.mock.patch(
        "apport.fileutils.allowed_to_report", MagicMock(return_value=True)
    )
    def test_crash_details(self):
        """Crash report with showing details"""

        def show_details():
            if Gtk.events_pending():
                return True  # pragma: no cover
            if not self.app.w("show_details").get_visible():
                return True  # pragma: no cover
            self.app.w("show_details").clicked()
            GLib.timeout_add(self.POLLING_INTERVAL_MS, cont)
            return False

        def cont():
            if Gtk.events_pending():
                return True  # pragma: no cover
            # wait until data collection is done and tree filled
            if self.app.tree_model.get_iter_first() is None:
                return True  # pragma: no cover

            self.assertTrue(self.app.w("continue_button").get_visible())
            self.app.w("continue_button").clicked()
            return False

        GLib.timeout_add(self.POLLING_INTERVAL_MS, show_details)
        info_collection_window = self.app.w("window_information_collection")
        with unittest.mock.patch.object(
            info_collection_window, "show", wraps=info_collection_window.show
        ) as info_collection_window_show_mock:
            self.app.run_crash(self.app.report_file)

        # we should have reported one crash
        self.assertEqual(self.app.crashdb.latest_id(), 0)
        r = self.app.crashdb.download(0)
        self.assertEqual(r["ProblemType"], "Crash")
        self.assertEqual(r["ExecutablePath"], "/bin/bash")

        # we already collected details, do not show the progress dialog again
        info_collection_window_show_mock.assert_not_called()

        # data was collected
        self.assertTrue(r["Package"].startswith("bash "))
        self.assertIn("libc", r["Dependencies"])

        # upload dialog shown
        self.assertEqual(self.app.ui_start_upload_progress.call_count, 1)
        self.assertEqual(self.app.ui_stop_upload_progress.call_count, 1)

        # URL was opened
        self.assertEqual(self.app.open_url.call_count, 1)

    @unittest.mock.patch.object(GTKUserInterface, "open_url", MagicMock())
    @unittest.mock.patch.object(
        GTKUserInterface, "ui_start_upload_progress", MagicMock()
    )
    @unittest.mock.patch.object(
        GTKUserInterface, "ui_stop_upload_progress", MagicMock()
    )
    @unittest.mock.patch(
        "apport.fileutils.allowed_to_report", MagicMock(return_value=True)
    )
    def test_broken_crash_details(self):
        """Broken crash report with showing details."""
        # pylint: disable=attribute-defined-outside-init
        self.error_title = None
        self.error_text = None

        def show_details():
            if Gtk.events_pending():
                return True  # pragma: no cover
            if not self.app.w("show_details").get_visible():
                return True  # pragma: no cover
            self.app.w("show_details").clicked()
            GLib.timeout_add(self.POLLING_INTERVAL_MS, cont)
            return False

        def cont():
            # wait until data collection is done and tree filled
            if Gtk.events_pending():
                return True  # pragma: no cover
            if self.app.tree_model.get_iter_first() is None:
                return True  # pragma: no cover

            self.assertTrue(self.app.w("continue_button").get_visible())
            self.app.w("continue_button").clicked()
            GLib.timeout_add(self.POLLING_INTERVAL_MS, ack_error)
            return False

        def ack_error():
            # wait until error dialog gets visible
            if Gtk.events_pending():
                return True  # pragma: no cover
            if not self.app.md:
                return True  # pragma: no cover
            self.error_title = self.app.md.get_title()
            self.error_text = self.app.md.get_property("text")
            self.app.md.response(0)
            return False

        # damage core dump in report file
        with open(self.app.report_file, encoding="utf-8") as f:
            lines = f.readlines()
        lines[-1] = " iiiiiiiiiiiiAAAA\n"
        with open(self.app.report_file, "w", encoding="utf-8") as f:
            f.write("".join(lines))
        self.app.report = None
        GLib.timeout_add(self.POLLING_INTERVAL_MS, show_details)
        self.app.run_crash(self.app.report_file)

        # upload dialog not shown
        self.assertEqual(self.app.ui_start_upload_progress.call_count, 0)
        self.assertEqual(self.app.ui_stop_upload_progress.call_count, 0)

        # no URL was opened
        self.assertEqual(self.app.open_url.call_count, 0)

        # no crash uploaded
        self.assertEqual(self.app.crashdb.latest_id(), -1)

        # proper error message
        self.assertIsNotNone(self.error_title)
        self.assertIn("cannot be reported", self.error_text)
        self.assertIn("decompressing", self.error_text)

    @unittest.mock.patch.object(GTKUserInterface, "open_url", MagicMock())
    @unittest.mock.patch(
        "apport.fileutils.allowed_to_report", MagicMock(return_value=True)
    )
    def test_crash_noaccept(self):
        """Crash report with non-accepting crash DB"""

        def cont():
            if Gtk.events_pending():
                return True  # pragma: no cover
            if not self.app.w("continue_button").get_visible():
                return True  # pragma: no cover
            self.app.w("continue_button").clicked()
            return False

        GLib.timeout_add(self.POLLING_INTERVAL_MS, cont)
        self.app.crashdb.options["problem_types"] = ["bug"]
        info_collection_window = self.app.w("window_information_collection")
        with unittest.mock.patch.object(
            info_collection_window, "show", wraps=info_collection_window.show
        ) as info_collection_window_show_mock:
            self.app.run_crash(self.app.report_file)

        # we should not have reported the crash
        self.assertEqual(self.app.crashdb.latest_id(), -1)
        self.assertEqual(self.app.open_url.call_count, 0)

        # no progress dialog for non-accepting DB
        info_collection_window_show_mock.assert_not_called()

        # data was collected for whoopsie
        r = self.app.report
        self.assertEqual(r["ProblemType"], "Crash")
        self.assertEqual(r["ExecutablePath"], "/bin/bash")
        self.assertTrue(r["Package"].startswith("bash "))
        self.assertIn("libc", r["Dependencies"])

    @unittest.mock.patch.object(GTKUserInterface, "open_url", MagicMock())
    @unittest.mock.patch(
        "apport.fileutils.allowed_to_report", MagicMock(return_value=True)
    )
    def test_kerneloops_nodetails(self):
        """Kernel oops report without showing details"""

        def cont():
            if Gtk.events_pending():
                return True  # pragma: no cover
            if not self.app.w("continue_button").get_visible():
                return True  # pragma: no cover
            self.app.w("continue_button").clicked()
            return False

        # remove the crash from setUp() and create a kernel oops
        os.remove(self.app.report_file)
        subprocess.run(
            [get_data_directory() / "kernel_oops"],
            check=True,
            input=b"Plasma conduit phase misalignment",
        )

        GLib.timeout_add(self.POLLING_INTERVAL_MS, cont)
        self.app.run_crashes()

        # we should have reported one crash
        self.assertEqual(self.app.crashdb.latest_id(), 0)
        r = self.app.crashdb.download(0)
        self.assertEqual(r["ProblemType"], "KernelOops")
        self.assertEqual(r["OopsText"], "Plasma conduit phase misalignment")

        # data was collected
        self.assertIn("linux", r["Package"])
        self.assertIn("Plasma conduit", r["Title"])
        if not r["Package"].endswith(" (not installed)"):
            self.assertIn("Dependencies", r)

        # URL was opened
        self.assertEqual(self.app.open_url.call_count, 1)

    def test_bug_report_installed_package(self):
        """Bug report for installed package"""

        def c():
            if Gtk.events_pending():
                return True  # pragma: no cover
            dont_send_button = self.app.w("dont_send_button")
            if not self.has_click_event_connected(dont_send_button):
                return True  # pragma: no cover
            dont_send_button.clicked()
            return False

        self.app.report_file = None
        self.app.args.package = "bash"
        GLib.timeout_add(self.POLLING_INTERVAL_MS, c)
        self.app.run_report_bug()

        self.assertEqual(self.app.report["ProblemType"], "Bug")
        self.assertEqual(self.app.report["SourcePackage"], "bash")
        self.assertTrue(self.app.report["Package"].startswith("bash "))
        self.assertNotEqual(self.app.report["Dependencies"], "")

    def test_bug_report_uninstalled_package(self):
        """Bug report for uninstalled package"""

        def c():
            if Gtk.events_pending():
                return True  # pragma: no cover
            dont_send_button = self.app.w("dont_send_button")
            if not self.has_click_event_connected(dont_send_button):
                return True  # pragma: no cover
            dont_send_button.clicked()
            return False

        pkg = apport.packaging.get_uninstalled_package()
        self.app.report_file = None
        self.app.args.package = pkg
        GLib.timeout_add(self.POLLING_INTERVAL_MS, c)
        self.app.run_report_bug()

        self.assertEqual(self.app.report["ProblemType"], "Bug")
        self.assertEqual(
            self.app.report["SourcePackage"], apport.packaging.get_source(pkg)
        )
        self.assertEqual(self.app.report["Package"], f"{pkg} (not installed)")

    @unittest.mock.patch.object(GTKUserInterface, "open_url", MagicMock())
    def test_update_report(self):
        """Updating an existing report."""
        self.app.report_file = None

        def cont():
            if Gtk.events_pending():
                return True  # pragma: no cover
            if self.app.tree_model.get_iter_first() is None:
                return True  # pragma: no cover
            self.app.w("continue_button").clicked()
            return False

        # upload empty report
        crash_id = self.app.crashdb.upload({})
        self.assertEqual(crash_id, 0)
        self.app.args.update_report = 0
        self.app.args.package = "bash"

        GLib.timeout_add(self.POLLING_INTERVAL_MS, cont)
        self.app.run_update_report()

        # no new bug reported
        self.assertEqual(self.app.crashdb.latest_id(), 0)

        # bug was updated
        r = self.app.crashdb.download(0)
        self.assertTrue(r["Package"].startswith("bash "))
        self.assertIn("libc", r["Dependencies"])
        self.assertIn("DistroRelease", r)

        # No URL in this mode
        self.assertEqual(self.app.open_url.call_count, 0)

    @unittest.mock.patch.object(GTKUserInterface, "open_url", MagicMock())
    def test_update_report_different_binary_source(self):
        """Updating an existing report on a source package which does not have
        a binary of the same name"""
        self.app.report_file = None

        def cont():
            if Gtk.events_pending():
                return True  # pragma: no cover
            if self.app.tree_model.get_iter_first() is None:
                return True  # pragma: no cover
            self.app.w("continue_button").clicked()
            return False

        # this test assumes that the source package name is not an
        # installed binary package
        source_pkg = "shadow"
        self.assertRaises(ValueError, apport.packaging.get_version, source_pkg)

        # create source package hook, as otherwise there is nothing to collect
        with open(
            os.path.join(self.hook_dir, f"source_{source_pkg}.py"),
            "w",
            encoding="utf-8",
        ) as f:
            f.write('def add_info(r, ui):\n r["MachineType"]="Laptop"\n')

        # upload empty report
        crash_id = self.app.crashdb.upload({})
        self.assertEqual(crash_id, 0)

        # run in update mode for that bug
        self.app.args.update_report = 0
        self.app.args.package = source_pkg

        GLib.timeout_add(self.POLLING_INTERVAL_MS, cont)
        self.app.run_update_report()

        # no new bug reported
        self.assertEqual(self.app.crashdb.latest_id(), 0)

        # bug was updated
        r = self.app.crashdb.download(0)
        self.assertIn("ProcEnviron", r)
        self.assertIn("DistroRelease", r)
        self.assertIn("Uname", r)
        self.assertEqual(r["MachineType"], "Laptop")

        # No URL in this mode
        self.assertEqual(self.app.open_url.call_count, 0)

    @unittest.mock.patch.object(GTKUserInterface, "get_desktop_entry", MagicMock())
    def test_missing_icon(self):
        # LP: 937354
        self.app.report["ProblemType"] = "Crash"
        self.app.report["Package"] = "apport 1.2.3~0ubuntu1"
        self.app.get_desktop_entry.return_value = {
            "name": "apport",
            "icon": "nonexistent",
        }
        GLib.idle_add(Gtk.main_quit)
        self.app.ui_present_report_details(True)

    def test_resizing(self):
        """Problem report window resizability and sizing."""

        def show_details(data):
            if not self.app.w("show_details").get_visible():
                return True  # pragma: no cover

            data["orig_size"] = self.app.w("dialog_crash_new").get_size()
            data["orig_resizable"] = self.app.w("dialog_crash_new").get_resizable()
            self.app.w("show_details").clicked()
            GLib.timeout_add(200, hide_details, data)
            return False

        def hide_details(data):
            # wait until data collection is done and tree filled
            if self.app.tree_model.get_iter_first() is None:
                return True  # pragma: no cover

            data["detail_size"] = self.app.w("dialog_crash_new").get_size()
            data["detail_resizable"] = self.app.w("dialog_crash_new").get_resizable()
            self.app.w("show_details").clicked()
            GLib.timeout_add(200, details_hidden, data)
            return False

        def details_hidden(data):
            # wait until data collection is done and tree filled
            if self.app.w("details_scrolledwindow").get_visible():
                return True  # pragma: no cover

            data["hidden_size"] = self.app.w("dialog_crash_new").get_size()
            data["hidden_resizable"] = self.app.w("dialog_crash_new").get_resizable()
            Gtk.main_quit()
            return False

        data = {}
        GLib.timeout_add(200, show_details, data)
        self.app.run_crash(self.app.report_file)

        # when showing details, dialog should get considerably bigger
        self.assertGreater(data["detail_size"][1], data["orig_size"][1] + 100)

        # when hiding details, original size should be restored
        self.assertEqual(data["orig_size"], data["hidden_size"])

        # should only be resizable in details mode
        self.assertFalse(data["orig_resizable"])
        self.assertTrue(data["detail_resizable"])
        self.assertFalse(data["hidden_resizable"])

    def test_dialog_nonascii(self):
        """Non-ASCII title/text in dialogs"""

        def close(response):
            if Gtk.events_pending():
                return True  # pragma: no cover
            if not self.app.md:
                return True  # pragma: no cover
            self.app.md.response(response)
            return False

        # unicode arguments
        GLib.timeout_add(self.POLLING_INTERVAL_MS, close, 0)
        self.app.ui_info_message(
            b"title \xe2\x99\xaa".decode("UTF-8"), b"text \xe2\x99\xaa".decode("UTF-8")
        )

        # with URLs
        GLib.timeout_add(self.POLLING_INTERVAL_MS, close, 0)
        self.app.ui_info_message(
            "title", b"http://example.com \xe2\x99\xaa".decode("UTF-8")
        )

    def test_immediate_close(self):
        """Close details window immediately."""
        # this reproduces https://launchpad.net/bugs/938090
        self.app.w("dialog_crash_new").destroy()
        GLib.idle_add(Gtk.main_quit)
        self.app.run_crash(self.app.report_file)

    @unittest.mock.patch.object(
        GTKUserInterface, "ui_start_upload_progress", MagicMock()
    )
    def test_close_during_collect(self):
        """Close details window during information collection"""

        def show_details():
            if Gtk.events_pending():
                return True  # pragma: no cover
            if not self.app.w("show_details").get_visible():
                return True  # pragma: no cover
            self.app.w("show_details").clicked()
            GLib.timeout_add(self.POLLING_INTERVAL_MS, close)
            return False

        def close():
            if Gtk.events_pending():
                return True  # pragma: no cover
            self.app.w("dialog_crash_new").destroy()
            return False

        GLib.timeout_add(self.POLLING_INTERVAL_MS, show_details)
        self.app.run_crash(self.app.report_file)

        self.assertEqual(self.app.ui_start_upload_progress.call_count, 0)

    def test_text_to_markup(self):
        """Test text_to_markup() with different URLs."""
        urls = ["https://example.com", "https://example.com/file-bug-report"]
        for url in urls:
            with self.subTest(url=url):
                self.assertEqual(
                    apport_gtk.text_to_markup(f"Contact them via {url} for help."),
                    f'Contact them via <a href="{url}">{url}</a> for help.',
                )

    def test_text_to_markup_url_followed_by_dot(self):
        """Test text_to_markup() with https URL followed by a dot."""
        url = "https://example.com/file-bug-report"
        self.assertEqual(
            apport_gtk.text_to_markup(f"Für Hilfe gehen Sie über {url}."),
            f'Für Hilfe gehen Sie über <a href="{url}">{url}</a>.',
        )

    def test_ui_run_terminal(self) -> None:
        """Test ui_run_terminal."""
        self.assertTrue(
            self.app.ui_has_terminal(), "installed terminal application needed"
        )
        self.app.ui_run_terminal("true")

    def test_ui_update_view_destroyed(self):
        """Test ui_update_view if the dialog is already destroyed."""
        self.app.w("details_treeview").destroy()
        self.app.ui_update_view()

    @staticmethod
    def has_click_event_connected(widget):
        signal_id = GObject.signal_lookup("clicked", widget)
        signal_handler_id = GObject.signal_handler_find(
            widget,
            GObject.SignalMatchType.ID | GObject.SignalMatchType.UNBLOCKED,
            signal_id,
            0,
            None,
            0,
            0,
        )
        return signal_handler_id != 0



#!/usr/bin/python3

"""Installer script for Apport."""

# TODO: Address following pylint complaints
# pylint: disable=invalid-name

import glob
import logging
import os.path
import subprocess
import sys

from setuptools_apport.java import register_java_sub_commands

try:
    import DistUtilsExtra.auto
    from DistUtilsExtra.command.build_extra import build_extra
except ImportError:
    sys.stderr.write(
        "To build Apport you need https://launchpad.net/python-distutils-extra\n"
    )
    sys.exit(1)

BASH_COMPLETIONS = "share/bash-completion/completions/"


class clean_java_subdir(DistUtilsExtra.auto.clean_build_tree):
    """Java crash handler clean command."""

    def run(self):
        DistUtilsExtra.auto.clean_build_tree.run(self)
        for root, _, files in os.walk("java"):
            for f in files:
                if f.endswith(".jar") or f.endswith(".class"):
                    os.unlink(os.path.join(root, f))


class install_fix_completion_symlinks(DistUtilsExtra.auto.install_auto):
    """Fix symlinks in the bash completion dir."""

    def run(self):
        log = logging.getLogger(__name__)
        DistUtilsExtra.auto.install_auto.run(self)

        autoinstalled_completion_dir = os.path.join(
            self.install_data, "share", "apport", "bash-completion"
        )
        for completion in glob.glob("data/bash-completion/*"):
            try:
                source = os.readlink(completion)
            except OSError:
                continue
            dest = os.path.join(
                self.install_data, BASH_COMPLETIONS, os.path.basename(completion)
            )
            if not os.path.exists(dest):
                continue

            log.info("Convert %s into a symlink to %s...", dest, source)
            os.remove(dest)
            os.symlink(source, dest)

            autoinstalled = os.path.join(
                autoinstalled_completion_dir, os.path.basename(completion)
            )
            os.remove(autoinstalled)

        # Clean-up left-over bash-completion from auto install
        if os.path.isdir(autoinstalled_completion_dir):
            os.rmdir(autoinstalled_completion_dir)


#
# main
#

from apport.ui import __version__  # noqa: E402, pylint: disable=C0413

# determine systemd unit directory
try:
    systemd_unit_dir = subprocess.check_output(
        ["pkg-config", "--variable=systemdsystemunitdir", "systemd"],
        universal_newlines=True,
    ).strip()
    systemd_tmpfiles_dir = subprocess.check_output(
        ["pkg-config", "--variable=tmpfilesdir", "systemd"], universal_newlines=True
    ).strip()
except (FileNotFoundError, subprocess.CalledProcessError):
    # hardcoded fallback path
    systemd_unit_dir = "/lib/systemd/system"
    systemd_tmpfiles_dir = "/usr/lib/tmpfiles.d"

try:
    udev_dir = subprocess.check_output(
        ["pkg-config", "--variable=udevdir", "udev"], text=True
    ).strip()
except (FileNotFoundError, subprocess.CalledProcessError):
    udev_dir = "/lib/udev"

cmdclass = register_java_sub_commands(build_extra, install_fix_completion_symlinks)
DistUtilsExtra.auto.setup(
    name="apport",
    author="Martin Pitt",
    author_email="martin.pitt@ubuntu.com",
    url="https://launchpad.net/apport",
    license="gpl",
    description="intercept, process, and report crashes and bug reports",
    packages=["apport", "apport.crashdb_impl", "apport.packaging_impl"],
    version=__version__,
    data_files=[
        ("share/doc/apport/", glob.glob("doc/*.txt")),
        # these are not supposed to be called directly, use apport-bug instead
        ("share/apport", ["gtk/apport-gtk", "kde/apport-kde"]),
        (BASH_COMPLETIONS, glob.glob("data/bash-completion/*")),
        ("lib/pm-utils/sleep.d/", glob.glob("pm-utils/sleep.d/*")),
        (f"{udev_dir}/rules.d", glob.glob("udev/*.rules")),
        (
            systemd_unit_dir,
            glob.glob("data/systemd/*.service") + glob.glob("data/systemd/*.socket"),
        ),
        (
            f"{systemd_unit_dir}/systemd-coredump@.service.d",
            ["data/systemd/systemd-coredump@.service.d/apport-coredump-hook.conf"],
        ),
        (systemd_tmpfiles_dir, glob.glob("data/systemd/*.conf")),
    ],
    cmdclass={
        "build": build_extra,
        "clean": clean_java_subdir,
        "install": install_fix_completion_symlinks,
    }
    | cmdclass,
)



"""Test test helper functions. Test inception for the win!"""

import os
import sys
import unittest

from tests.helper import pidof, read_shebang


class TestHelper(unittest.TestCase):
    # pylint: disable=missing-class-docstring,missing-function-docstring

    def test_pidof_non_existing_program(self):
        self.assertEqual(pidof("non-existing"), set())

    def test_pidof_running_python(self):
        pids = pidof(sys.executable)
        self.assertGreater(len(pids), 0)
        self.assertIn(os.getpid(), pids)

    def test_read_shebang_binary(self):
        self.assertIsNone(read_shebang(sys.executable))

    def test_read_shebang_shell_script(self):
        self.assertEqual(read_shebang("/usr/bin/ldd"), "/bin/bash")
